import torch
import numpy as np


from typing import Optional


def create_meshgrid(
    height: int,
    width: int,
    normalized_coordinates: bool = True,
    device: Optional[torch.device] = torch.device("cpu"),
    dtype: torch.dtype = torch.float32,
) -> torch.Tensor:
    """Generate a coordinate grid for an image.
    When the flag ``normalized_coordinates`` is set to True, the grid is
    normalized to be in the range :math:`[-1,1]` to be consistent with the pytorch
    function :py:func:`torch.nn.functional.grid_sample`.
    Args:
        height: the image height (rows).
        width: the image width (cols).
        normalized_coordinates: whether to normalize
          coordinates in the range :math:`[-1,1]` in order to be consistent with the
          PyTorch function :py:func:`torch.nn.functional.grid_sample`.
        device: the device on which the grid will be generated.
        dtype: the data type of the generated grid.
    Return:
        grid tensor with shape :math:`(1, H, W, 2)`.
    Example:
        >>> create_meshgrid(2, 2)
        tensor([[[[-1., -1.],
                  [ 1., -1.]],
        <BLANKLINE>
                 [[-1.,  1.],
                  [ 1.,  1.]]]])
        >>> create_meshgrid(2, 2, normalized_coordinates=False)
        tensor([[[[0., 0.],
                  [1., 0.]],
        <BLANKLINE>
                 [[0., 1.],
                  [1., 1.]]]])
    """
    xs: torch.Tensor = torch.linspace(0, width - 1, width, device=device, dtype=dtype)
    ys: torch.Tensor = torch.linspace(0, height - 1, height, device=device, dtype=dtype)
    # Fix TracerWarning
    # Note: normalize_pixel_coordinates still gots TracerWarning since new width and height
    #       tensors will be generated.
    # Below is the code using normalize_pixel_coordinates:
    # base_grid: torch.Tensor = torch.stack(torch.meshgrid([xs, ys], indexing='ij'), dim=2)
    # if normalized_coordinates:
    #     base_grid = K.geometry.normalize_pixel_coordinates(base_grid, height, width)
    # return torch.unsqueeze(base_grid.transpose(0, 1), dim=0)
    if normalized_coordinates:
        xs = (xs / (width - 1) - 0.5) * 2
        ys = (ys / (height - 1) - 0.5) * 2
    # generate grid by stacking coordinates
    base_grid: torch.Tensor = torch.stack(
        torch.meshgrid([xs, ys], indexing="ij")
    ).transpose(
        1, 2
    )  # 2xHxW
    return torch.unsqueeze(base_grid, dim=0).permute(0, 2, 3, 1)  # 1xHxWx2


def backwarp(img, flow):
    _, _, H, W = img.size()

    u = flow[:, 0, :, :]
    v = flow[:, 1, :, :]

    gridX, gridY = np.meshgrid(np.arange(W), np.arange(H))
    gridX = torch.tensor(
        gridX,
        requires_grad=False,
    ).to(flow.device)
    gridY = torch.tensor(
        gridY,
        requires_grad=False,
    ).to(flow.device)
    x = gridX.unsqueeze(0).expand_as(u).float() + u
    y = gridY.unsqueeze(0).expand_as(v).float() + v
    # range -1 to 1
    x = 2 * (x / W - 0.5)
    y = 2 * (y / H - 0.5)
    # stacking X and Y
    grid = torch.stack((x, y), dim=3)
    # Sample pixels using bilinear interpolation.
    imgOut = torch.nn.functional.grid_sample(img, grid, align_corners=False)

    return imgOut


@torch.no_grad()
def compute_supervision_coarse(flow, occlusions, scale: int):
    N, _, H, W = flow.shape
    Hc, Wc = int(np.ceil(H / scale)), int(np.ceil(W / scale))

    occlusions_c = occlusions[:, :, ::scale, ::scale]
    flow_c = flow[:, :, ::scale, ::scale] / scale
    occlusions_c = occlusions_c.reshape(N, Hc * Wc)

    grid_c = (
        create_meshgrid(Hc, Wc, False, device=flow.device)
        .reshape(1, Hc * Wc, 2)
        .repeat(N, 1, 1)
    )
    warp_c = grid_c + flow_c.permute(0, 2, 3, 1).reshape(N, Hc * Wc, 2)
    warp_c = warp_c.round().long()

    def out_bound_mask(pt, w, h):
        return (
            (pt[..., 0] < 0) + (pt[..., 0] >= w) + (pt[..., 1] < 0) + (pt[..., 1] >= h)
        )

    occlusions_c[out_bound_mask(warp_c, Wc, Hc)] = 1
    warp_c = warp_c[..., 0] + warp_c[..., 1] * Wc

    b_ids, i_ids = torch.split(torch.nonzero(occlusions_c == 0), 1, dim=1)
    conf_matrix_gt = torch.zeros(N, Hc * Wc, Hc * Wc, device=flow.device)
    j_ids = warp_c[b_ids, i_ids]
    conf_matrix_gt[b_ids, i_ids, j_ids] = 1

    return conf_matrix_gt


def compute_coarse_loss(conf, conf_gt, cfg):
    c_pos_w, c_neg_w = cfg.POS_WEIGHT, cfg.NEG_WEIGHT
    pos_mask, neg_mask = conf_gt == 1, conf_gt == 0

    if cfg.COARSE_TYPE == "cross_entropy":
        conf = torch.clamp(conf, 1e-6, 1 - 1e-6)
        loss_pos = -torch.log(conf[pos_mask])
        loss_neg = -torch.log(1 - conf[neg_mask])

        return c_pos_w * loss_pos.mean() + c_neg_w * loss_neg.mean()
    elif cfg.COARSE_TYPE == "focal":
        conf = torch.clamp(conf, 1e-6, 1 - 1e-6)
        alpha = cfg.FOCAL_ALPHA
        gamma = cfg.FOCAL_GAMMA
        loss_pos = (
            -alpha * torch.pow(1 - conf[pos_mask], gamma) * (conf[pos_mask]).log()
        )
        loss_neg = (
            -alpha * torch.pow(conf[neg_mask], gamma) * (1 - conf[neg_mask]).log()
        )
        return c_pos_w * loss_pos.mean() + c_neg_w * loss_neg.mean()
    else:
        raise ValueError("Unknown coarse loss: {type}".format(type=cfg.COARSE_TYPE))


def compute_fine_loss(kflow, kflow_gt, cfg):
    fine_correct_thr = cfg.WINDOW_SIZE // 2 * 2
    error = (kflow - kflow_gt).abs()
    correct = torch.max(error, dim=1)[0] < fine_correct_thr
    rate = torch.sum(correct).float() / correct.shape[0]
    num = correct.shape[0]
    return error[correct].mean(), rate.item(), num


def compute_flow_loss(flow, flow_gt):
    loss = (flow - flow_gt).abs().mean()
    epe = torch.sum((flow - flow_gt) ** 2, dim=1).sqrt()

    metrics = {
        "epe": epe.mean().item(),
        "1px": (epe < 1).float().mean().item(),
        "3px": (epe < 3).float().mean().item(),
        "5px": (epe < 5).float().mean().item(),
    }

    return loss, metrics



import torch

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )


class ConvBlock(torch.nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride, padding):
        super(ConvBlock, self).__init__()

        self.conv = torch.nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            padding_mode="zeros",
            bias=True,
        )
        self.relu = torch.nn.LeakyReLU(negative_slope=0.1, inplace=False)

    def forward(self, x):
        return self.relu(self.conv(x))


class Refine(torch.nn.Module):
    def __init__(self, feature_dim, patch_size, num_layers):
        super(Refine, self).__init__()

        self.patch_size = patch_size

        self.correlation_sampler = SpatialCorrelationSampler(
            kernel_size=1, patch_size=patch_size, stride=1, padding=0, dilation=1
        )

        self.conv1 = ConvBlock(
            patch_size**2 + feature_dim + 2, 96, kernel_size=3, stride=1, padding=1
        )

        self.conv_layers = torch.nn.ModuleList(
            [
                ConvBlock(96, 96, kernel_size=3, stride=1, padding=1)
                for i in range(num_layers)
            ]
        )

        self.conv2 = ConvBlock(96, 64, kernel_size=3, stride=1, padding=1)
        self.conv3 = ConvBlock(64, 32, kernel_size=3, stride=1, padding=1)
        self.conv4 = torch.nn.Conv2d(
            32, 2, kernel_size=3, stride=1, padding=1, padding_mode="zeros", bias=True
        )

    def forward(self, feature_0, feature_1, flow_0):
        b, c, h, w = feature_0.shape

        attn = self.correlation_sampler(feature_0, feature_1).view(b, -1, h, w)
        # attn = F.softmax(attn, dim=1)

        x = torch.cat([attn, feature_0, flow_0], dim=1)

        x = self.conv1(x)

        for layer in self.conv_layers:
            x = layer(x)

        x = self.conv2(x)
        x = self.conv3(x)

        return self.conv4(x)



"""Perform spatial correlation to generate a cost volume.

This is meant to be a simplified version of the SpatialCorrelationSampler from
https://github.com/ClementPinard/Pytorch-Correlation-extension.

This version is implemented purely in PyTorch. However, it only supports correlation with 1x1 kernels.
It is also not as efficient as the original SpatialCorrelationSampler.
"""

# =============================================================================
# Copyright 2021 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math
from typing import Optional, Tuple, Union

from einops import rearrange
import torch
import torch.nn as nn
import torch.nn.functional as F


def iter_spatial_correlation_sample(
    input1: torch.Tensor,
    input2: torch.Tensor,
    kernel_size: Union[int, Tuple[int, int]] = 1,
    patch_size: Union[int, Tuple[int, int]] = 1,
    stride: Union[int, Tuple[int, int]] = 1,
    padding: Union[int, Tuple[int, int]] = 0,
    dilation: Union[int, Tuple[int, int]] = 1,
    dilation_patch: Union[int, Tuple[int, int]] = 1,
) -> torch.Tensor:
    """Apply spatial correlation sampling from input1 to input2 using iteration in PyTorch.

    This docstring is taken and adapted from the original package.

    Every parameter except input1 and input2 can be either single int or a pair of int. For more information about
    Spatial Correlation Sampling, see this page. https://lmb.informatik.uni-freiburg.de/Publications/2015/DFIB15/

    Parameters
    ----------
    input1 : torch.Tensor
        The origin feature map.
    input2 : torch.Tensor
        The target feature map.
    kernel_size : Union[int, Tuple[int, int]], default 1
        Total size of your correlation kernel, in pixels
    patch_size : Union[int, Tuple[int, int]], default 1
        Total size of your patch, determining how many different shifts will be applied.
    stride : Union[int, Tuple[int, int]], default 1
        Stride of the spatial sampler, will modify output height and width.
    padding : Union[int, Tuple[int, int]], default 0
        Padding applied to input1 and input2 before applying the correlation sampling, will modify output height and width.
    dilation : Union[int, Tuple[int, int]], default 1
        Similar to dilation in convolution.
    dilation_patch : Union[int, Tuple[int, int]], default 1
        Step for every shift in patch.

    Returns
    -------
    torch.Tensor
        Result of correlation sampling.

    Raises
    ------
    NotImplementedError
        If kernel_size != 1.
    NotImplementedError
        If dilation != 1.
    """
    # Make inputs be tuples
    kernel_size = (
        (kernel_size, kernel_size) if isinstance(kernel_size, int) else kernel_size
    )
    patch_size = (patch_size, patch_size) if isinstance(patch_size, int) else patch_size
    stride = (stride, stride) if isinstance(stride, int) else stride
    padding = (padding, padding) if isinstance(padding, int) else padding
    dilation = (dilation, dilation) if isinstance(dilation, int) else dilation
    dilation_patch = (
        (dilation_patch, dilation_patch)
        if isinstance(dilation_patch, int)
        else dilation_patch
    )

    if kernel_size[0] != 1 or kernel_size[1] != 1:
        raise NotImplementedError("Only kernel_size=1 is supported.")
    if dilation[0] != 1 or dilation[1] != 1:
        raise NotImplementedError("Only dilation=1 is supported.")

    if max(padding) > 0:
        input1 = F.pad(input1, (padding[1], padding[1], padding[0], padding[0]))
        input2 = F.pad(input2, (padding[1], padding[1], padding[0], padding[0]))

    input2 = F.pad(
        input2,
        (
            dilation_patch[1] * ((patch_size[1] - 1) // 2),
            dilation_patch[1] * (patch_size[1] // 2),
            dilation_patch[0] * ((patch_size[0] - 1) // 2),
            dilation_patch[0] * (patch_size[0] // 2),
        ),
    )

    b, _, h, w = input1.shape
    input1 = input1[:, :, :: stride[0], :: stride[1]]
    sh, sw = input1.shape[2:4]
    corr = torch.zeros(b, patch_size[0], patch_size[1], sh, sw).to(
        dtype=input1.dtype, device=input1.device
    )

    for i in range(0, patch_size[0] * dilation_patch[0], dilation_patch[0]):
        for j in range(0, patch_size[1] * dilation_patch[1], dilation_patch[1]):
            p2 = input2[:, :, i : i + h, j : j + w]
            p2 = p2[:, :, :: stride[0], :: stride[1]]
            corr[:, i // dilation_patch[0], j // dilation_patch[1]] = (input1 * p2).sum(
                dim=1
            )

    return corr


class IterSpatialCorrelationSampler(nn.Module):
    """Apply spatial correlation sampling from two inputs using iteration in PyTorch."""

    def __init__(
        self,
        kernel_size: Union[int, Tuple[int, int]] = 1,
        patch_size: Union[int, Tuple[int, int]] = 1,
        stride: Union[int, Tuple[int, int]] = 1,
        padding: Union[int, Tuple[int, int]] = 0,
        dilation: Union[int, Tuple[int, int]] = 1,
        dilation_patch: Union[int, Tuple[int, int]] = 1,
    ) -> None:
        """Initialize IterSpatialCorrelationSampler.

        Parameters
        ----------
        kernel_size : Union[int, Tuple[int, int]], default 1
            Total size of your correlation kernel, in pixels
        patch_size : Union[int, Tuple[int, int]], default 1
            Total size of your patch, determining how many different shifts will be applied.
        stride : Union[int, Tuple[int, int]], default 1
            Stride of the spatial sampler, will modify output height and width.
        padding : Union[int, Tuple[int, int]], default 0
            Padding applied to input1 and input2 before applying the correlation sampling, will modify output height and width.
        dilation : Union[int, Tuple[int, int]], default 1
            Similar to dilation in convolution.
        dilation_patch : Union[int, Tuple[int, int]], default 1
            Step for every shift in patch.
        """
        super(IterSpatialCorrelationSampler, self).__init__()
        self.kernel_size = kernel_size
        self.patch_size = patch_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.dilation_patch = dilation_patch

    def forward(self, input1: torch.Tensor, input2: torch.Tensor) -> torch.Tensor:
        """Compute the correlation sampling from input1 to input2.

        Parameters
        ----------
        input1 : torch.Tensor
            The origin feature map.
        input2 : torch.Tensor
            The target feature map.

        Returns
        -------
        torch.Tensor
            Result of correlation sampling.
        """
        return iter_spatial_correlation_sample(
            input1=input1,
            input2=input2,
            kernel_size=self.kernel_size,
            patch_size=self.patch_size,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            dilation_patch=self.dilation_patch,
        )


def _init_coords_grid(flow: torch.Tensor) -> torch.Tensor:
    """Creates a grid of absolute 2D coordinates.

    Parameters
    ----------
    flow : torch.Tensor
        The optical flow field to translate the points from input1. The flow values should be represented in number of pixels
        (do not provide normalized values, e.g. between -1 and 1). It should be a 4D tensor (b, 2, h, w), where
        flow[:, 0] represent the horizontal flow and flow[:, 1] the vertical ones.

    Returns
    -------
    torch.Tensor
        The grid with the 2D coordinates of the pixels.
    """
    b, _, h, w = flow.shape
    coords_grid = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")
    coords_grid = torch.stack(coords_grid[::-1], dim=0).to(
        dtype=flow.dtype, device=flow.device
    )
    coords_grid = coords_grid[None].repeat(b, 1, 1, 1)
    return coords_grid


def iter_translated_spatial_correlation_sample(
    input1: torch.Tensor,
    input2: torch.Tensor,
    flow: Optional[torch.Tensor] = None,
    coords: Optional[torch.Tensor] = None,
    kernel_size: Union[int, Tuple[int, int]] = 1,
    patch_size: Union[int, Tuple[int, int]] = 1,
    stride: Union[int, Tuple[int, int]] = 1,
    padding: Union[int, Tuple[int, int]] = 0,
    dilation: Union[int, Tuple[int, int]] = 1,
    dilation_patch: Union[int, Tuple[int, int]] = 1,
    coords_grid: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """Apply spatial correlation sampling with translation from input1 to input2 using iteration in PyTorch.

    This operation is equivalent to first translating the points from input1 using the given flow, and then doing a local
    correlation sampling around the translated points.

    This allows us to do correlation sampling without warping the second input.

    Every parameter except input1, input2, and flow can be either single int or a pair of int. For more information about
    Spatial Correlation Sampling (without translation), see this page: https://lmb.informatik.uni-freiburg.de/Publications/2015/DFIB15/

    Parameters
    ----------
    input1 : torch.Tensor
        The origin feature map.
    input2 : torch.Tensor
        The target feature map.
    flow : Optional[torch.Tensor]
        This argument and "coords" are mutually exclusive, only one of them can be not None.
        The optical flow field to translate the points from input1. The flow values should be represented in number of pixels
        (do not provide normalized values, e.g. between -1 and 1). It should be a 4D tensor (b, 2, h, w), where
        flow[:, 0] represent the horizontal flow and flow[:, 1] the vertical ones.
    coords : torch.Tensor
        This argument and "flow" are mutually exclusive, only one of them can be not None.
        This value should be equivalent to "flow" + "coords_grid".
    kernel_size : Union[int, Tuple[int, int]], default 1
        Total size of your correlation kernel, in pixels
    patch_size : Union[int, Tuple[int, int]], default 1
        Total size of your patch, determining how many different shifts will be applied.
    stride : Union[int, Tuple[int, int]], default 1
        Stride of the spatial sampler, will modify output height and width.
    padding : Union[int, Tuple[int, int]], default 0
        Padding applied to input1 and input2 before applying the correlation sampling, will modify output height and width.
    dilation : Union[int, Tuple[int, int]], default 1
        Similar to dilation in convolution.
    dilation_patch : Union[int, Tuple[int, int]], default 1
        Step for every shift in patch.
    coords_grid : Optional[torch.Tensor], default None
        A tensor with the same shape as flow containing a grid of 2D coordinates of the pixels. This can be created using torch.meshgrid.
        This parameter is optional. If not provided, the grid will be created internally. Only useful if the grid can be buffered somewhere
        to avoid recreating it at every call.

    Returns
    -------
    torch.Tensor
        Result of correlation sampling.

    Raises
    ------
    NotImplementedError
        If kernel_size != 1.
    NotImplementedError
        If dilation != 1.
    """
    assert (flow is None and coords is not None) or (
        flow is not None and coords is None
    )
    # Make inputs be tuples
    kernel_size = (
        (kernel_size, kernel_size) if isinstance(kernel_size, int) else kernel_size
    )
    patch_size = (patch_size, patch_size) if isinstance(patch_size, int) else patch_size
    stride = (stride, stride) if isinstance(stride, int) else stride
    padding = (padding, padding) if isinstance(padding, int) else padding
    dilation = (dilation, dilation) if isinstance(dilation, int) else dilation
    dilation_patch = (
        (dilation_patch, dilation_patch)
        if isinstance(dilation_patch, int)
        else dilation_patch
    )

    if kernel_size[0] != 1 or kernel_size[1] != 1:
        raise NotImplementedError("Only kernel_size=1 is supported.")
    if dilation[0] != 1 or dilation[1] != 1:
        raise NotImplementedError("Only dilation=1 is supported.")

    if max(padding) > 0:
        input1 = F.pad(input1, (padding[1], padding[1], padding[0], padding[0]))
        input2 = F.pad(input2, (padding[1], padding[1], padding[0], padding[0]))

    b, _, h, w = input2.shape
    input1 = input1[:, :, :: stride[0], :: stride[1]]
    sh, sw = input1.shape[2:4]
    corr = torch.zeros(b, patch_size[0], patch_size[1], sh, sw).to(
        dtype=input1.dtype, device=input1.device
    )

    if coords is None:
        if coords_grid is None:
            coords_grid = _init_coords_grid(flow)
        coords = coords_grid + flow

    cx = 2 * coords[:, 0] / (w - 1) - 1
    cy = 2 * coords[:, 1] / (h - 1) - 1

    offset = (
        dilation_patch[0] * ((patch_size[0] - 1) // 2),
        dilation_patch[1] * ((patch_size[1] - 1) // 2),
    )

    for i in range(0, patch_size[0] * dilation_patch[0], dilation_patch[0]):
        for j in range(0, patch_size[1] * dilation_patch[1], dilation_patch[1]):
            grid = torch.stack(
                [
                    cx + 2 * (j - offset[1]) / float(w - 1),
                    cy + 2 * (i - offset[0]) / float(h - 1),
                ],
                dim=-1,
            )
            p2 = F.grid_sample(input2, grid, mode="bilinear", align_corners=True)
            p2 = p2[:, :, :: stride[0], :: stride[1]]
            corr[:, i // dilation_patch[0], j // dilation_patch[1]] = (input1 * p2).sum(
                dim=1
            )

    return corr


class IterTranslatedSpatialCorrelationSampler(nn.Module):
    """Apply translated spatial correlation sampling from two inputs using iteration in PyTorch.

    This operation is equivalent to first translating the points from input1 using the given flow, and then doing a local
    correlation sampling around the translated points.

    This allows us to do correlation sampling without warping the second input.
    """

    def __init__(
        self,
        kernel_size: Union[int, Tuple[int, int]] = 1,
        patch_size: Union[int, Tuple[int, int]] = 1,
        stride: Union[int, Tuple[int, int]] = 1,
        padding: Union[int, Tuple[int, int]] = 0,
        dilation: Union[int, Tuple[int, int]] = 1,
        dilation_patch: Union[int, Tuple[int, int]] = 1,
    ) -> None:
        """Initialize IterTranslatedSpatialCorrelationSampler.

        Parameters
        ----------
        kernel_size : Union[int, Tuple[int, int]], default 1
            Total size of your correlation kernel, in pixels
        patch_size : Union[int, Tuple[int, int]], default 1
            Total size of your patch, determining how many different shifts will be applied.
        stride : Union[int, Tuple[int, int]], default 1
            Stride of the spatial sampler, will modify output height and width.
        padding : Union[int, Tuple[int, int]], default 0
            Padding applied to input1 and input2 before applying the correlation sampling, will modify output height and width.
        dilation : Union[int, Tuple[int, int]], default 1
            Similar to dilation in convolution.
        dilation_patch : Union[int, Tuple[int, int]], default 1
            Step for every shift in patch.
        """
        super(IterTranslatedSpatialCorrelationSampler, self).__init__()
        self.kernel_size = kernel_size
        self.patch_size = patch_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.dilation_patch = dilation_patch

        self.coords_grid = None

    def forward(
        self, input1: torch.Tensor, input2: torch.Tensor, flow: torch.Tensor
    ) -> torch.Tensor:
        """Compute the correlation sampling from input1 to input2.

        Parameters
        ----------
        input1 : torch.Tensor
            The origin feature map.
        input2 : torch.Tensor
            The target feature map.
        flow : torch.Tensor
            The optical flow field to translate the points from input1. The flow values should be represented in number of pixels
            (do not provide normalized values, e.g. between -1 and 1). It should be a 4D tensor (b, 2, h, w), where
            flow[:, 0] represent the horizontal flow and flow[:, 1] the vertical ones.

        Returns
        -------
        torch.Tensor
            Result of correlation sampling.
        """
        b, _, h, w = flow.shape
        if (
            self.coords_grid is None
            or self.coords_grid.shape[2] != h
            or self.coords_grid.shape[3] != w
        ):
            self.coords_grid = _init_coords_grid(flow)
        if self.coords_grid.shape[0] != b:
            self.coords_grid = self.coords_grid[:1].repeat(b, 1, 1, 1)

        return iter_translated_spatial_correlation_sample(
            input1=input1,
            input2=input2,
            flow=flow,
            kernel_size=self.kernel_size,
            patch_size=self.patch_size,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            dilation_patch=self.dilation_patch,
            coords_grid=self.coords_grid,
        )


class IterativeCorrBlock(nn.Module):
    """Another wrapper for iter_translated_spatial_correlation_sample.

    This block is designed to mimic the operations of RAFT's AlternateCorrBlock package (see ptlflow/models/raft/corr.py).
    This block can be used when alt_cuda_corr has not been compiled (see ptlflow/utils/external/alt_cuda_corr).

    IMPORTANT: this implementation is slower than alt_cuda_corr.
    """

    def __init__(
        self,
        fmap1: torch.Tensor,
        fmap2: torch.Tensor,
        radius: int = 1,
        num_levels: int = 1,
    ):
        """Initialize IterativeCorrBlock.

        Parameters
        ----------
        fmap1 : torch.Tensor
            The origin feature map.
        fmap2 : torch.Tensor
            The target feature map.
        radius : int, default 1
            The radius if the correlation patch. The patch_size will be 2 * radius + 1.
        num_levels : int, default 1
            Number of correlation pooling levels to use (see ptlflow/models/raft/corr.py).
        """
        super(IterativeCorrBlock, self).__init__()

        self.patch_size = 2 * radius + 1
        self.num_levels = num_levels

        self.pyramid = [(fmap1, fmap2)]
        for _ in range(self.num_levels):
            fmap1 = F.avg_pool2d(fmap1, 2, stride=2)
            fmap2 = F.avg_pool2d(fmap2, 2, stride=2)
            self.pyramid.append((fmap1, fmap2))

    def forward(self, coords):
        """Compute the correlation sampling from input1 to input2.

        Parameters
        ----------
        coords : torch.Tensor
            The addition (optical flow + coords_grid) to translate the points from input1. The coords values should be represented in number of pixels
            (do not provide normalized values, e.g. between -1 and 1). It should be a 4D tensor (b, 2, h, w), where
            coords[:, 0] represent the x axis and flow[:, 1] the y axis.

        Returns
        -------
        torch.Tensor
            Result of correlation sampling.
        """
        dim = self.pyramid[0][0].shape[1]

        corr_list = []
        for i in range(self.num_levels):
            fmap1_i = self.pyramid[0][0]
            fmap2_i = self.pyramid[i][1]

            coords_i = coords / 2**i
            corr = iter_translated_spatial_correlation_sample(
                input1=fmap1_i,
                input2=fmap2_i,
                coords=coords_i,
                patch_size=self.patch_size,
            )
            corr = rearrange(corr, "b c d h w -> b (d c) h w")
            corr_list.append(corr)

        corr = torch.cat(corr_list, dim=1)
        return corr / math.sqrt(dim)



from typing import Optional

import torch.nn.functional as F

from ptlflow.utils.registry import register_model
from .FlowFormer.encoders import twins_svt_large, convnext_large
from .FlowFormer.PerCostFormer3.encoder import MemoryEncoder
from .FlowFormer.PerCostFormer3.decoder import MemoryDecoder
from .FlowFormer.PerCostFormer3.cnn import BasicEncoder
from .utils import compute_grid_indices, compute_weight
from ..base_model.base_model import BaseModel


class FlowFormerPlusPlus(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformerplusplus-chairs-a7745dd5.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformerplusplus-things-4db3ecff.ckpt",
        "things288960": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformerplusplus-things_288960-a4291d41.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformerplusplus-sintel-d14a1968.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/flowformerplusplus-kitti-65b828c3.ckpt",
    }

    def __init__(
        self,
        cnet: str = "twins",
        fnet: str = "twins",
        pretrain: bool = False,
        patch_size: int = 8,
        cost_heads_num: int = 1,
        cost_latent_input_dim: int = 64,
        cost_latent_token_num: int = 8,
        cost_latent_dim: int = 128,
        pe: str = "linear",
        encoder_depth: int = 3,
        encoder_latent_dim: int = 256,
        decoder_depth: int = 32,
        dropout: float = 0.0,
        vert_c_dim: int = 64,
        query_latent_dim: int = 64,
        cost_encoder_res: bool = True,
        pic_size: tuple[int, int, int, int] = (368, 496, 368, 496),
        del_layers: bool = True,
        pretrain_mode: bool = False,
        use_convertor: bool = False,
        patch_embed: str = "single",
        cross_attn: str = "all",
        droppath: float = 0.0,
        vertical_encoder_attn: str = "twins",
        use_patch: bool = False,
        fix_pe: bool = False,
        gt_r: int = 15,
        flow_or_pe: str = "and",
        no_sc: bool = False,
        r_16: int = -1,
        quater_refine: bool = False,
        use_rpe: bool = False,
        gma: str = "GMA",
        detach_local: bool = False,
        use_tile_input: bool = True,
        tile_height: int = 432,
        tile_sigma: float = 0.05,
        train_size: Optional[tuple[int, int]] = None,
        crop_cost_volume: bool = False,
        mask_ratio: float = 0.0,
        attn_dim: int = 0,
        expand_factor: float = 4.0,
        query_num: int = 0,
        no_border: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(output_stride=32, loss_fn=None, **kwargs)

        self.cnet = cnet
        self.fnet = fnet
        self.cost_encoder_res = cost_encoder_res
        self.cost_heads_num = cost_heads_num
        self.cost_latent_dim = cost_latent_dim
        self.cost_latent_input_dim = cost_latent_input_dim
        self.cost_latent_token_num = cost_latent_token_num
        self.decoder_depth = decoder_depth
        self.dropout = dropout
        self.encoder_depth = encoder_depth
        self.encoder_latent_dim = encoder_latent_dim
        self.gma = gma
        self.patch_size = patch_size
        self.pe = pe
        self.pretrain = pretrain
        self.query_latent_dim = query_latent_dim
        self.vert_c_dim = vert_c_dim
        self.use_tile_input = use_tile_input
        self.tile_height = tile_height
        self.tile_sigma = tile_sigma
        self.train_size = train_size
        self.pic_size = pic_size
        self.del_layers = del_layers
        self.pretrain_mode = pretrain_mode
        self.use_convertor = use_convertor
        self.patch_embed = patch_embed
        self.cross_attn = cross_attn
        self.droppath = droppath
        self.vertical_encoder_attn = vertical_encoder_attn
        self.use_patch = use_patch
        self.fix_pe = fix_pe
        self.gt_r = gt_r
        self.flow_or_pe = flow_or_pe
        self.no_sc = no_sc
        self.r_16 = r_16
        self.quater_refine = quater_refine
        self.use_rpe = use_rpe
        self.detach_local = detach_local
        self.crop_cost_volume = crop_cost_volume
        self.mask_ratio = mask_ratio
        self.attn_dim = attn_dim
        self.expand_factor = expand_factor
        self.query_num = query_num
        self.no_border = no_border

        H1, W1, H2, W2 = pic_size
        H_offset = (H1 - H2) // 2
        W_offset = (W1 - W2) // 2

        self.memory_encoder = MemoryEncoder(
            cost_heads_num=cost_heads_num,
            use_convertor=use_convertor,
            r_16=r_16,
            crop_cost_volume=crop_cost_volume,
            del_layers=del_layers,
            H_offset=H_offset,
            W_offset=W_offset,
            fnet=fnet,
            pretrain=pretrain,
            pretrain_mode=pretrain_mode,
            encoder_latent_dim=encoder_latent_dim,
            vertical_encoder_attn=vertical_encoder_attn,
            cost_latent_token_num=cost_latent_token_num,
            cost_latent_dim=cost_latent_dim,
            cost_encoder_res=cost_encoder_res,
            mask_ratio=mask_ratio,
            patch_size=patch_size,
            cost_latent_input_dim=cost_latent_input_dim,
            pe=pe,
            encoder_depth=encoder_depth,
            cross_attn=cross_attn,
            dropout=dropout,
            vert_c_dim=vert_c_dim,
            patch_embed=patch_embed,
            use_rpe=use_rpe,
            droppath=droppath,
            attn_dim=attn_dim,
            expand_factor=expand_factor,
        )
        self.memory_decoder = MemoryDecoder(
            gma=gma,
            use_patch=use_patch,
            detach_local=detach_local,
            use_rpe=use_rpe,
            r_16=r_16,
            quater_refine=quater_refine,
            fix_pe=fix_pe,
            gt_r=gt_r,
            query_num=query_num,
            no_border=no_border,
            W_offset=W_offset,
            H_offset=H_offset,
            query_latent_dim=query_latent_dim,
            cost_latent_input_dim=cost_latent_input_dim,
            cost_heads_num=cost_heads_num,
            encoder_latent_dim=encoder_latent_dim,
            decoder_depth=decoder_depth,
            cost_latent_dim=cost_latent_dim,
            patch_size=patch_size,
            flow_or_pe=flow_or_pe,
            dropout=dropout,
            pe=pe,
            no_sc=no_sc,
        )
        if cnet == "twins":
            self.context_encoder = twins_svt_large(
                pretrained=self.pretrain, del_layers=del_layers
            )
        elif cnet == "basicencoder":
            self.context_encoder = BasicEncoder(output_dim=256, norm_fn="instance")
        elif cnet == "convnext":
            self.context_encoder = convnext_large(pretrained=self.pretrain)

        if pretrain_mode:
            print("[In pretrain mode, freeze context encoder]")
            for param in self.context_encoder.parameters():
                param.requires_grad = False

        self.showed_warning = False

    def forward(self, inputs, mask=None, output=None):
        """Estimate optical flow between pair of frames"""
        if self.pretrain_mode:
            image1 = (image1 + 1) * 127.5
            image2 = (image2 + 1) * 127.5
            loss = self.pretrain_forward(image1, image2, mask=mask, output=output)
            return loss

        if self.train_size is not None:
            train_size = self.train_size
        else:
            train_size = self.train_size

        if self.use_tile_input and train_size is None and not self.showed_warning:
            print(
                "WARNING: --train_size is not provided and it cannot be loaded from the checkpoint either. Flowformer++ will run without input tile."
            )
            self.showed_warning = True

        if self.use_tile_input and train_size is not None:
            return self.forward_tile(inputs, train_size)
        else:
            return self.forward_pad(inputs)

    def forward_pad(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        prev_flow = None
        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            prev_flow = inputs["prev_preds"]["flow_small"]

        flow_predictions, flow_small = self.predict(
            images[:, 0], images[:, 1], prev_flow
        )
        output_flow = flow_predictions[-1]

        if self.training:
            for i, p in enumerate(flow_predictions):
                flow_predictions[i] = self.postprocess_predictions(
                    p, image_resizer, is_flow=True
                )
            outputs = {
                "flows": flow_predictions[-1][:, None],
                "flow_preds": flow_predictions,
            }
        else:
            output_flow = self.postprocess_predictions(
                output_flow, image_resizer, is_flow=True
            )
            outputs = {"flows": output_flow[:, None], "flow_small": flow_small}

        return outputs

    def forward_tile(self, inputs, train_size):
        input_size = inputs["images"].shape[-2:]
        image_size = (max(self.tile_height, input_size[-2]), input_size[-1])
        hws = compute_grid_indices(image_size, train_size)
        device = inputs["images"].device
        weights = compute_weight(
            hws, image_size, train_size, self.tile_sigma, device=device
        )

        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            target_size=image_size,
            pad_two_side=False,
            pad_mode="constant",
            pad_value=-1,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        flows = 0
        flow_count = 0

        for idx, (h, w) in enumerate(hws):
            image1_tile = image1[:, :, h : h + train_size[0], w : w + train_size[1]]
            image2_tile = image2[:, :, h : h + train_size[0], w : w + train_size[1]]

            flow_predictions, _ = self.predict(image1_tile, image2_tile)
            flow_pre = flow_predictions[-1]

            padding = (
                w,
                image_size[1] - w - train_size[1],
                h,
                image_size[0] - h - train_size[0],
                0,
                0,
            )
            flows += F.pad(flow_pre * weights[idx], padding)
            flow_count += F.pad(weights[idx], padding)

        output_flow = flows / flow_count

        output_flow = self.postprocess_predictions(
            output_flow, image_resizer, is_flow=True
        )
        return {"flows": output_flow[:, None]}

    def predict(self, image1, image2, prev_flow=None):
        """Estimate optical flow between pair of frames"""
        data = {}

        context, _ = self.context_encoder(image1)
        context_quater = None

        (
            cost_memory,
            cost_patches,
            feat_s_quater,
            feat_t_quater,
        ) = self.memory_encoder(image1, image2, data, context)

        flow_predictions, flow_small = self.memory_decoder(
            cost_memory,
            context,
            context_quater,
            feat_s_quater,
            feat_t_quater,
            data,
            prev_flow=prev_flow,
            cost_patches=cost_patches,
        )

        return flow_predictions, flow_small


@register_model
class flowformer_pp(FlowFormerPlusPlus):
    pass



def torch_init_model(model, total_dict, key, rank=0):
    if key in total_dict:
        state_dict = total_dict[key]
    else:
        state_dict = total_dict
    missing_keys = []
    unexpected_keys = []
    error_msgs = []
    # copy state_dict so _load_from_state_dict can modify it
    metadata = getattr(state_dict, "_metadata", None)
    state_dict = state_dict.copy()
    if metadata is not None:
        state_dict._metadata = metadata

    def load(module, prefix=""):
        local_metadata = {} if metadata is None else metadata.get(prefix[:-1], {})
        module._load_from_state_dict(
            state_dict=state_dict,
            prefix=prefix,
            local_metadata=local_metadata,
            strict=True,
            missing_keys=missing_keys,
            unexpected_keys=unexpected_keys,
            error_msgs=error_msgs,
        )
        for name, child in module._modules.items():
            if child is not None:
                load(child, prefix + name + ".")

    load(model, prefix="")

    if rank == 0:
        print("missing keys:{}".format(missing_keys))
        print("unexpected keys:{}".format(unexpected_keys))
        print("error msgs:{}".format(error_msgs))



import torch
from .kv_memory_store import KeyValueMemoryStore
from .MemFlowNet.memory_util import *


class MemoryManager:
    """
    Manages all three memory stores and the transition between working/long-term memory
    """

    def __init__(
        self,
        train_avg_length,
        enable_long_term,
        enable_long_term_count_usage,
        top_k,
        max_mid_term_frames,
        min_mid_term_frames,
    ):
        self.train_avg_length = train_avg_length

        self.enable_long_term = enable_long_term
        self.enable_long_term_usage = enable_long_term_count_usage
        # top_k for softmax
        self.top_k = top_k
        self.max_mt_frames = max_mid_term_frames
        self.min_mt_frames = min_mid_term_frames

        # dimensions will be inferred from input later
        self.CK = self.CV = None
        self.H = self.W = None

        self.work_mem = KeyValueMemoryStore(count_usage=self.enable_long_term)
        self.reset_config = True

    def _readout(self, affinity, v):
        # this function is for a single object group
        return v @ affinity

    def match_memory(self, query_key, current_key, current_value, scale):
        # query_key: B x C^k x H x W
        h, w = query_key.shape[-2:]

        query_key = query_key.flatten(start_dim=2)
        if current_key is not None and current_value is not None:
            current_key = current_key.flatten(start_dim=2)
            current_value = current_value.flatten(start_dim=2)

        """
        Memory readout using keys
        """
        if self.work_mem.engaged():
            # No long-term memory
            if current_key is not None and current_value is not None:
                memory_key = torch.cat([self.work_mem.key, current_key], -1)
            else:
                memory_key = self.work_mem.key

            scale = scale * math.log(memory_key.shape[-1], self.train_avg_length)
            similarity = (
                torch.einsum("b c l, b c t -> b t l", query_key, memory_key) * scale
            )

            affinity, usage = do_softmax(
                similarity, inplace=True, top_k=self.top_k, return_usage=True
            )

            if current_key is not None and current_value is not None:
                all_memory_value = torch.cat([self.work_mem.value, current_value], -1)
                work_usage = usage[:, : -h * w]
            else:
                all_memory_value = self.work_mem.value
                work_usage = usage
            # Record memory usage for working memory
            self.work_mem.update_usage(work_usage.flatten())
            # else:
            #     raise NotImplementedError
        else:
            # No working-term memory
            if current_key is not None and current_value is not None:
                memory_key = current_key
                scale = scale * math.log(memory_key.shape[-1], self.train_avg_length)
                similarity = (
                    torch.einsum("b c l, b c t -> b t l", query_key, memory_key) * scale
                )
                affinity = do_softmax(
                    similarity, inplace=True, top_k=None, return_usage=False
                )
                all_memory_value = current_value
            else:
                return 0

        # Shared affinity within each group
        all_readout_mem = self._readout(affinity, all_memory_value)

        return all_readout_mem.view(all_readout_mem.shape[0], -1, h, w)

    def add_memory(self, key, value):
        # key: 1*C*H*W
        # value: 1*C*H*W
        if self.H is None or self.reset_config:
            self.reset_config = False
            self.H, self.W = key.shape[-2:]
            self.HW = self.H * self.W
            self.min_work_elements = self.min_mt_frames * self.HW
            self.max_work_elements = self.max_mt_frames * self.HW

        # key:   1*C*N
        # value: 1*C*N
        key = key.flatten(start_dim=2)
        value = value.flatten(start_dim=2)

        self.CK = key.shape[1]
        self.CV = value.shape[1]

        self.work_mem.add(key, value)

        # Do memory compressed if needed
        if self.work_mem.size >= self.max_work_elements:
            # Remove obsolete features if needed
            self.compress_features()

    def compress_features(self):
        # remove consolidated working memory
        self.work_mem.sieve_by_range(0, -self.min_work_elements)



import torch
import torch.nn as nn
import torch.nn.functional as F

from .update import GMAUpdateBlock
from ..encoders import twins_svt_large, convnext_Xlarge_4x, convnext_base_2x
from .corr import CorrBlock, OLCorrBlock, AlternateCorrBlock
from ...utils.utils import bilinear_sampler, coords_grid, upflow8
from .gma import Attention, Aggregate

from torchvision.utils import save_image

autocast = torch.cuda.amp.autocast


class MOFNet(nn.Module):
    def __init__(self, cfg):
        super().__init__()
        self.cfg = cfg

        self.hidden_dim = hdim = self.cfg.feat_dim // 2
        self.context_dim = cdim = self.cfg.feat_dim // 2

        cfg.corr_radius = 4

        # feature network, context network, and update block
        if cfg.cnet == "twins":
            print("[Using twins as context encoder]")
            self.cnet = twins_svt_large(pretrained=self.cfg.pretrain)
        elif cfg.cnet == "basicencoder":
            print("[Using basicencoder as context encoder]")
            self.cnet = BasicEncoder(output_dim=256, norm_fn="instance")
        elif cfg.cnet == "convnext_Xlarge_4x":
            print("[Using convnext_Xlarge_4x as context encoder]")
            self.cnet = convnext_Xlarge_4x(pretrained=self.cfg.pretrain)
        elif cfg.cnet == "convnext_base_2x":
            print("[Using convnext_base_2x as context encoder]")
            self.cnet = convnext_base_2x(pretrained=self.cfg.pretrain)

        if cfg.fnet == "twins":
            print("[Using twins as feature encoder]")
            self.fnet = twins_svt_large(pretrained=self.cfg.pretrain)
        elif cfg.fnet == "basicencoder":
            print("[Using basicencoder as feature encoder]")
            self.fnet = BasicEncoder(output_dim=256, norm_fn="instance")
        elif cfg.fnet == "convnext_Xlarge_4x":
            print("[Using convnext_Xlarge_4x as feature encoder]")
            self.fnet = convnext_Xlarge_4x(pretrained=self.cfg.pretrain)
        elif cfg.fnet == "convnext_base_2x":
            print("[Using convnext_base_2x as feature encoder]")
            self.fnet = convnext_base_2x(pretrained=self.cfg.pretrain)

        hidden_dim_ratio = 256 // cfg.feat_dim

        if self.cfg.Tfusion == "stack":
            print("[Using stack.]")
            self.cfg.cost_heads_num = 1
            from .stack import SKUpdateBlock6_Deep_nopoolres_AllDecoder2

            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder2(
                args=self.cfg, hidden_dim=128 // hidden_dim_ratio
            )
        # elif self.cfg.Tfusion == 'resstack':
        #     print("[Using resstack.]")
        #     self.cfg.cost_heads_num = 1
        #     from .resstack import SKUpdateBlock6_Deep_nopoolres_AllDecoder2
        #     self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder2(args=self.cfg, hidden_dim=128)
        # elif self.cfg.Tfusion == 'stackcat':
        #     print("[Using stackcat.]")
        #     self.cfg.cost_heads_num = 1
        #     from .stackcat import SKUpdateBlock6_Deep_nopoolres_AllDecoder2
        #     self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder2(args=self.cfg, hidden_dim=128)

        print("[Using corr_fn {}]".format(self.cfg.corr_fn))

        gma_down_ratio = 256 // cfg.feat_dim

        self.att = Attention(
            args=self.cfg,
            dim=128 // hidden_dim_ratio,
            heads=1,
            max_pos_size=160,
            dim_head=128 // hidden_dim_ratio,
        )

        if self.cfg.context_3D:
            print("[Using 3D Conv on context feature.]")
            self.context_3D = nn.Sequential(
                nn.Conv3d(256, 256, 3, stride=1, padding=1),
                nn.GELU(),
                nn.Conv3d(256, 256, 3, stride=1, padding=1),
                nn.GELU(),
                nn.Conv3d(256, 256, 3, stride=1, padding=1),
                nn.GELU(),
            )

    def initialize_flow(self, img, bs, down_ratio):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(bs, H // down_ratio, W // down_ratio).to(img.device)
        coords1 = coords_grid(bs, H // down_ratio, W // down_ratio).to(img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def upsample_flow_4x(self, flow, mask):
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 4, 4, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(4 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 4 * H, 4 * W)

    def upsample_flow_2x(self, flow, mask):
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 2, 2, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(2 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 2 * H, 2 * W)

    def forward(self, images, data={}, flow_init=None):
        down_ratio = self.cfg.down_ratio

        B, N, _, H, W = images.shape

        images = 2 * (images / 255.0) - 1.0

        hdim = self.hidden_dim
        cdim = self.context_dim

        with autocast(enabled=self.cfg.mixed_precision):
            fmaps = self.fnet(images.reshape(B * N, 3, H, W)).reshape(
                B, N, -1, H // down_ratio, W // down_ratio
            )

        if self.cfg.corr_fn == "default":
            corr_fn = CorrBlock
        elif self.cfg.corr_fn == "efficient":
            corr_fn = AlternateCorrBlock
        forward_corr_fn = corr_fn(
            fmaps[:, 1 : N - 1, ...].reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            ),
            fmaps[:, 2:N, ...].reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            ),
            num_levels=self.cfg.corr_levels,
            radius=self.cfg.corr_radius,
        )
        backward_corr_fn = corr_fn(
            fmaps[:, 1 : N - 1, ...].reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            ),
            fmaps[:, 0 : N - 2, ...].reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            ),
            num_levels=self.cfg.corr_levels,
            radius=self.cfg.corr_radius,
        )

        with autocast(enabled=self.cfg.mixed_precision):
            cnet = self.cnet(images[:, 1 : N - 1, ...].reshape(B * (N - 2), 3, H, W))
            if self.cfg.context_3D:
                # print("!@!@@#!@#!@")
                cnet = cnet.reshape(B, N - 2, -1, H // 2, W // 2).permute(0, 2, 1, 3, 4)
                cnet = self.context_3D(cnet) + cnet
                # print(cnet.shape)
                cnet = cnet.permute(0, 2, 1, 3, 4).reshape(
                    B * (N - 2), -1, H // down_ratio, W // down_ratio
                )
            net, inp = torch.split(cnet, [hdim, cdim], dim=1)
            net = torch.tanh(net)
            inp = torch.relu(inp)
            attention = self.att(inp)

        forward_coords1, forward_coords0 = self.initialize_flow(
            images[:, 0, ...], bs=B * (N - 2), down_ratio=down_ratio
        )
        backward_coords1, backward_coords0 = self.initialize_flow(
            images[:, 0, ...], bs=B * (N - 2), down_ratio=down_ratio
        )

        flow_predictions = []  # forward flows followed by backward flows

        motion_hidden_state = None

        for itr in range(self.cfg.decoder_depth):
            forward_coords1 = forward_coords1.detach()
            backward_coords1 = backward_coords1.detach()

            forward_corr = forward_corr_fn(forward_coords1)
            backward_corr = backward_corr_fn(backward_coords1)

            forward_flow = forward_coords1 - forward_coords0
            backward_flow = backward_coords1 - backward_coords0

            with autocast(enabled=self.cfg.mixed_precision):
                net, motion_hidden_state, up_mask, delta_flow = self.update_block(
                    net,
                    motion_hidden_state,
                    inp,
                    forward_corr,
                    backward_corr,
                    forward_flow,
                    backward_flow,
                    forward_coords0,
                    attention,
                    bs=B,
                )

            forward_up_mask, backward_up_mask = torch.split(
                up_mask, [down_ratio**2 * 9, down_ratio**2 * 9], dim=1
            )

            forward_coords1 = forward_coords1 + delta_flow[:, 0:2, ...]
            backward_coords1 = backward_coords1 + delta_flow[:, 2:4, ...]

            # upsample predictions
            if down_ratio == 4:
                forward_flow_up = self.upsample_flow_4x(
                    forward_coords1 - forward_coords0, forward_up_mask
                )
                backward_flow_up = self.upsample_flow_4x(
                    backward_coords1 - backward_coords0, backward_up_mask
                )
            elif down_ratio == 2:
                forward_flow_up = self.upsample_flow_2x(
                    forward_coords1 - forward_coords0, forward_up_mask
                )
                backward_flow_up = self.upsample_flow_2x(
                    backward_coords1 - backward_coords0, backward_up_mask
                )
            elif down_ratio == 8:
                forward_flow_up = self.upsample_flow(
                    forward_coords1 - forward_coords0, forward_up_mask
                )
                backward_flow_up = self.upsample_flow(
                    backward_coords1 - backward_coords0, backward_up_mask
                )

            flow_predictions.append(
                torch.cat(
                    [
                        forward_flow_up.reshape(B, N - 2, 2, H, W),
                        backward_flow_up.reshape(B, N - 2, 2, H, W),
                    ],
                    dim=1,
                )
            )

        if self.training:
            return flow_predictions
        else:
            return flow_predictions[-1], flow_predictions[-1]



import torch
import torch.nn as nn
import torch.nn.functional as F
from .gma import Aggregate

from ...utils.utils import bilinear_sampler


class PCBlock4_Deep_nopool_res(nn.Module):
    def __init__(self, C_in, C_out, k_conv):
        super().__init__()
        self.conv_list = nn.ModuleList(
            [
                nn.Conv2d(
                    C_in, C_in, kernel, stride=1, padding=kernel // 2, groups=C_in
                )
                for kernel in k_conv
            ]
        )

        self.ffn1 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_in, 1, padding=0),
        )
        self.pw = nn.Conv2d(C_in, C_in, 1, padding=0)
        self.ffn2 = nn.Sequential(
            nn.Conv2d(C_in, int(1.5 * C_in), 1, padding=0),
            nn.GELU(),
            nn.Conv2d(int(1.5 * C_in), C_out, 1, padding=0),
        )

    def forward(self, x):
        x = F.gelu(x + self.ffn1(x))
        for conv in self.conv_list:
            x = F.gelu(x + conv(x))
        x = F.gelu(x + self.pw(x))
        x = self.ffn2(x)
        return x


class velocity_update_block(nn.Module):
    def __init__(self, C_in=43 + 128 + 43, C_out=43, C_hidden=64):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Conv2d(C_in, C_hidden, 3, padding=1),
            nn.GELU(),
            nn.Conv2d(C_hidden, C_hidden, 3, padding=1),
            nn.GELU(),
            nn.Conv2d(C_hidden, C_out, 3, padding=1),
        )

    def forward(self, x):
        return self.mlp(x)


class SKMotionEncoder6_Deep_nopool_res(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.cor_planes = cor_planes = (
            (args.corr_radius * 2 + 1) ** 2 * args.cost_heads_num * args.corr_levels
        )
        self.convc1 = PCBlock4_Deep_nopool_res(cor_planes, 128, k_conv=args.k_conv)
        self.convc2 = PCBlock4_Deep_nopool_res(256, 192, k_conv=args.k_conv)

        self.init_hidden_state = nn.Parameter(torch.randn(1, 1, 48, 1, 1))

        self.convf1_ = nn.Conv2d(4, 128, 1, 1, 0)
        self.convf2 = PCBlock4_Deep_nopool_res(128, 64, k_conv=args.k_conv)

        self.conv = PCBlock4_Deep_nopool_res(
            64 + 192 + 48 * 3, 128 - 4 + 48, k_conv=args.k_conv
        )

        self.velocity_update_block = velocity_update_block()

    def sample_flo_feat(self, flow, feat):
        sampled_feat = bilinear_sampler(feat, flow.permute(0, 2, 3, 1))
        return sampled_feat

    def forward(
        self,
        motion_hidden_state,
        forward_flow,
        backward_flow,
        coords0,
        forward_corr,
        backward_corr,
        bs,
    ):
        BN, _, H, W = forward_flow.shape
        N = BN // bs

        if motion_hidden_state is None:
            # print("initialized as None")
            motion_hidden_state = self.init_hidden_state.repeat(bs, N, 1, H, W)
        else:
            # print("later iterations")
            motion_hidden_state = motion_hidden_state.reshape(bs, N, -1, H, W)

        forward_loc = forward_flow + coords0
        backward_loc = backward_flow + coords0

        forward_motion_hidden_state = torch.cat(
            [
                motion_hidden_state[:, 1:, ...],
                torch.zeros(bs, 1, 48, H, W).to(motion_hidden_state.device),
            ],
            dim=1,
        ).reshape(BN, -1, H, W)
        backward_motion_hidden_state = torch.cat(
            [
                torch.zeros(bs, 1, 48, H, W).to(motion_hidden_state.device),
                motion_hidden_state[:, : N - 1, ...],
            ],
            dim=1,
        ).reshape(BN, -1, H, W)

        forward_cor = self.convc1(forward_corr)
        backward_cor = self.convc1(backward_corr)
        cor = F.gelu(torch.cat([forward_cor, backward_cor], dim=1))
        cor = self.convc2(cor)

        flow = torch.cat([forward_flow, backward_flow], dim=1)
        flo = self.convf1_(flow)
        flo = self.convf2(flo)

        cor_flo = torch.cat(
            [
                cor,
                flo,
                forward_motion_hidden_state,
                backward_motion_hidden_state,
                motion_hidden_state.reshape(BN, -1, H, W),
            ],
            dim=1,
        )
        out = self.conv(cor_flo)

        out, motion_hidden_state = torch.split(out, [124, 48], dim=1)

        return torch.cat([out, flow], dim=1), motion_hidden_state


class SKUpdateBlock6_Deep_nopoolres_AllDecoder2(nn.Module):
    def __init__(self, args, hidden_dim):
        super().__init__()
        self.args = args

        args.k_conv = [1, 15]
        args.PCUpdater_conv = [1, 7]

        self.encoder = SKMotionEncoder6_Deep_nopool_res(args)
        self.gru = PCBlock4_Deep_nopool_res(
            128 + hidden_dim + hidden_dim + 128, 128, k_conv=args.PCUpdater_conv
        )
        self.flow_head = PCBlock4_Deep_nopool_res(128, 4, k_conv=args.k_conv)

        self.mask = nn.Sequential(
            nn.Conv2d(128, 256, 3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 64 * 9 * 2, 1, padding=0),
        )

        self.aggregator = Aggregate(args=self.args, dim=128, dim_head=128, heads=1)

    def forward(
        self,
        net,
        motion_hidden_state,
        inp,
        forward_corr,
        backward_corr,
        forward_flow,
        backward_flow,
        coords0,
        attention,
        bs,
    ):
        motion_features, motion_hidden_state = self.encoder(
            motion_hidden_state,
            forward_flow,
            backward_flow,
            coords0,
            forward_corr,
            backward_corr,
            bs=bs,
        )
        motion_features_global = self.aggregator(attention, motion_features)
        inp_cat = torch.cat([inp, motion_features, motion_features_global], dim=1)

        # Attentional update
        net = self.gru(torch.cat([net, inp_cat], dim=1))

        delta_flow = self.flow_head(net)

        # scale mask to balence gradients
        mask = 100.0 * self.mask(net)
        return net, motion_hidden_state, mask, delta_flow



import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from .update import BasicUpdateBlock
from .extractor import BasicEncoder
from .corr import CorrBlock, CorrBlock1D
from .cost_agg import CostAggregation
from .utils import coords_grid, upflow8
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""
        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            i_weight = self.gamma ** (n_predictions - i - 1)
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += i_weight * (valid * i_loss).mean()

        return flow_loss


class Guidance(nn.Module):
    def __init__(self, channels=32, refine=False):
        super(Guidance, self).__init__()
        self.bn_relu = nn.Sequential(nn.InstanceNorm2d(channels), nn.ReLU(inplace=True))
        self.conv0 = nn.Sequential(
            nn.Conv2d(3, 16, kernel_size=3, padding=1),
            nn.InstanceNorm2d(16),
            nn.ReLU(inplace=True),
            nn.Conv2d(16, int(channels / 4), kernel_size=3, stride=2, padding=1),
            nn.InstanceNorm2d(int(channels / 4)),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                int(channels / 4), int(channels / 2), kernel_size=3, stride=2, padding=1
            ),
            nn.InstanceNorm2d(int(channels / 2)),
            nn.ReLU(inplace=True),
            nn.Conv2d(int(channels / 2), channels, kernel_size=3, stride=2, padding=1),
            nn.InstanceNorm2d(channels),
            nn.ReLU(inplace=True),
        )
        inner_channels = channels // 4
        self.wsize = 20
        self.conv1 = nn.Sequential(
            nn.Conv2d(channels * 2, inner_channels, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(inner_channels, inner_channels, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                inner_channels, inner_channels, kernel_size=3, stride=1, padding=1
            ),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
        )
        self.conv3 = nn.Sequential(
            nn.Conv2d(inner_channels, inner_channels, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                inner_channels, inner_channels, kernel_size=3, stride=1, padding=1
            ),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
        )
        self.conv11 = nn.Sequential(
            nn.Conv2d(
                inner_channels, inner_channels * 2, kernel_size=3, stride=2, padding=1
            ),
            nn.InstanceNorm2d(inner_channels * 2),
            nn.ReLU(inplace=True),
        )
        self.conv12 = nn.Sequential(
            nn.Conv2d(
                inner_channels * 2,
                inner_channels * 2,
                kernel_size=3,
                stride=1,
                padding=1,
            ),
            nn.InstanceNorm2d(inner_channels * 2),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                inner_channels * 2,
                inner_channels * 2,
                kernel_size=3,
                stride=1,
                padding=1,
            ),
            nn.InstanceNorm2d(inner_channels * 2),
            nn.ReLU(inplace=True),
        )
        self.weights = nn.Sequential(
            nn.Conv2d(inner_channels, inner_channels, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(inner_channels, self.wsize, kernel_size=3, stride=1, padding=1),
        )
        self.weight_sg1 = nn.Sequential(
            nn.Conv2d(inner_channels, inner_channels, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                inner_channels, self.wsize * 2, kernel_size=3, stride=1, padding=1
            ),
        )
        self.weight_sg2 = nn.Sequential(
            nn.Conv2d(inner_channels, inner_channels, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                inner_channels, self.wsize * 2, kernel_size=3, stride=1, padding=1
            ),
        )
        self.weight_sg11 = nn.Sequential(
            nn.Conv2d(inner_channels * 2, inner_channels * 2, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels * 2),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                inner_channels * 2, self.wsize * 2, kernel_size=3, stride=1, padding=1
            ),
        )
        self.weight_sg12 = nn.Sequential(
            nn.Conv2d(inner_channels * 2, inner_channels * 2, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels * 2),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                inner_channels * 2, self.wsize * 2, kernel_size=3, stride=1, padding=1
            ),
        )
        self.weight_sg3 = nn.Sequential(
            nn.Conv2d(inner_channels, inner_channels, kernel_size=3, padding=1),
            nn.InstanceNorm2d(inner_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                inner_channels, self.wsize * 2, kernel_size=3, stride=1, padding=1
            ),
        )
        # self.getweights = nn.Sequential(GetFilters(radius=1),
        #                                nn.Conv2d(9, 20, kernel_size=1, stride=1, padding=0, bias=False))

    def forward(self, fea, img):
        x = self.conv0(img)
        x = torch.cat((self.bn_relu(fea), x), 1)
        x = self.conv1(x)
        rem = x
        x = self.conv2(x) + rem
        rem = x
        guid = self.weights(x)
        x = self.conv3(x) + rem
        sg1 = self.weight_sg1(x)
        sg1_u, sg1_v = torch.split(sg1, (self.wsize, self.wsize), dim=1)
        sg2 = self.weight_sg2(x)
        sg2_u, sg2_v = torch.split(sg2, (self.wsize, self.wsize), dim=1)
        sg3 = self.weight_sg3(x)
        sg3_u, sg3_v = torch.split(sg3, (self.wsize, self.wsize), dim=1)
        x = self.conv11(x)
        rem = x
        x = self.conv12(x) + rem
        sg11 = self.weight_sg11(x)
        sg11_u, sg11_v = torch.split(sg11, (self.wsize, self.wsize), dim=1)
        sg12 = self.weight_sg12(x)
        sg12_u, sg12_v = torch.split(sg12, (self.wsize, self.wsize), dim=1)
        guid_u = dict(
            [
                ("sg1", sg1_u),
                ("sg2", sg2_u),
                ("sg3", sg3_u),
                ("sg11", sg11_u),
                ("sg12", sg12_u),
            ]
        )
        guid_v = dict(
            [
                ("sg1", sg1_v),
                ("sg2", sg2_v),
                ("sg3", sg3_v),
                ("sg11", sg11_v),
                ("sg12", sg12_v),
            ]
        )
        return guid, guid_u, guid_v


class SeparableFlow(BaseModel):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/separableflow-things-31fe3b2d.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/separableflow-sintel-4c9a8c03.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/separableflow-kitti-c9395318.ckpt",
        "universal": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/separableflow-universal-87350d91.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0.0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        hidden_dim: int = 128,
        context_dim: int = 128,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=8, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.dropout = dropout
        self.gamma = gamma
        self.max_flow = max_flow
        self.iters = iters
        self.hidden_dim = hidden_dim
        self.context_dim = context_dim

        hdim = self.hidden_dim
        cdim = self.context_dim

        # feature network, context network, and update block
        self.fnet = BasicEncoder(output_dim=256, norm_fn="instance", dropout=dropout)
        self.cnet = BasicEncoder(
            output_dim=hdim + cdim, norm_fn="batch", dropout=dropout
        )
        self.update_block = BasicUpdateBlock(
            corr_levels=corr_levels, corr_radius=corr_radius, hidden_dim=hdim
        )
        self.guidance = Guidance(channels=256)
        self.cost_agg1 = CostAggregation(in_channel=8)
        self.cost_agg2 = CostAggregation(in_channel=8)

    def freeze_bn(self):
        count1, count2, count3 = 0, 0, 0
        for m in self.modules():
            if isinstance(m, nn.SyncBatchNorm):
                count1 += 1
                m.eval()
            if isinstance(m, nn.BatchNorm2d):
                count2 += 1
                m.eval()
            if isinstance(m, nn.BatchNorm3d):
                count3 += 1
                # print(m)
                m.eval()
        # print(count1, count2, count3)
        # print(m)

    def initialize_flow(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 8, W // 8).to(img.device)
        coords1 = coords_grid(N, H // 8, W // 8).to(img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        hdim = self.hidden_dim
        cdim = self.context_dim

        # run the feature network
        fmap1, fmap2 = self.fnet([image1, image2])

        fmap1 = fmap1.float()
        fmap2 = fmap2.float()
        guid, guid_u, guid_v = self.guidance(fmap1.detach(), image1)
        corr_fn = CorrBlock(fmap1, fmap2, guid, radius=self.corr_radius)

        cnet = self.cnet(image1)
        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)
        corr1, corr2 = corr_fn(None, sep=True)
        coords0, coords1 = self.initialize_flow(image1)
        if self.training:
            u0, u1, flow_u, corr1 = self.cost_agg1(
                corr1, guid_u, max_shift=384, is_ux=True
            )
            v0, v1, flow_v, corr2 = self.cost_agg2(
                corr2, guid_v, max_shift=384, is_ux=False
            )
            flow_init = torch.cat((flow_u, flow_v), dim=1)

            flow_predictions = []
            flow_predictions.append(torch.cat((u0, v0), dim=1))
            flow_predictions.append(torch.cat((u1, v1), dim=1))
            flow_predictions.append(flow_init)

        else:
            flow_u, corr1 = self.cost_agg1(corr1, guid_u, max_shift=384, is_ux=True)
            flow_v, corr2 = self.cost_agg2(corr2, guid_v, max_shift=384, is_ux=False)
            flow_init = torch.cat((flow_u, flow_v), dim=1)
        flow_init = F.interpolate(
            flow_init.detach() / 8.0,
            [cnet.shape[2], cnet.shape[3]],
            mode="bilinear",
            align_corners=True,
        )
        corr1d_fn = CorrBlock1D(corr1, corr2, radius=self.corr_radius)
        coords1 = coords1 + flow_init
        for itr in range(self.iters):
            coords1 = coords1.detach()

            corr = corr_fn(coords1)  # index correlation volume
            corr1, corr2 = corr1d_fn(coords1)  # index correlation volume
            flow = coords1 - coords0
            net, up_mask, delta_flow = self.update_block(
                net, inp, corr, corr1, corr2, flow
            )

            # F(t+1) = F(t) + \Delta(t)
            coords1 = coords1 + delta_flow

            # upsample predictions
            if up_mask is None:
                flow_up = upflow8(coords1 - coords0)
            else:
                flow_up = self.upsample_flow(coords1 - coords0, up_mask)
            flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
            if self.training:
                flow_predictions.append(flow_up)

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        else:
            outputs = {"flows": flow_up[:, None], "flow_small": coords1 - coords0}

        return outputs


@register_model
@trainable
class separableflow(SeparableFlow):
    pass



import torch
import torch.nn.functional as F

from ptlflow.utils.registry import register_model
from .attention import Attention
from .extractor import BasicEncoder
from .corr import CorrBlock
from .update import Update
from ..base_model.base_model import BaseModel

try:
    from .softsplat import FunctionSoftsplat as forward_warping
except ModuleNotFoundError:
    forward_warping = None


class SplatFlow(BaseModel):
    pretrained_checkpoints = {
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/splatflow-kitti-2aa8e145.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        dropout: float = 0.0,
        gamma: float = 0.8,
        max_flow: float = 400,
        iters: int = 32,
        alternate_corr: bool = False,
        fast_inference: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(output_stride=8, loss_fn=None, **kwargs)

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.dropout = dropout
        self.gamma = gamma
        self.max_flow = max_flow
        self.iters = iters
        self.alternate_corr = alternate_corr
        self.fast_inference = fast_inference

        self.hdim = self.cdim = 128
        self.fnet = BasicEncoder(output_dim=256, norm_fn="instance")
        self.cnet = BasicEncoder(output_dim=self.hdim + self.cdim, norm_fn="batch")
        self.att = Attention(dim=self.cdim, heads=1, dim_head=self.cdim)
        self.update = Update(hidden_dim=self.hdim)

        self.has_shown_warning = False

        if forward_warping is None:
            raise ModuleNotFoundError("No module named 'cupy'")

    def init_coord(self, fmap):
        f_shape = fmap.shape
        H, W = f_shape[-2:]
        y0, x0 = torch.meshgrid(
            torch.arange(H).to(device=fmap.device, dtype=fmap.dtype),
            torch.arange(W).to(device=fmap.device, dtype=fmap.dtype),
            indexing="ij",
        )
        coord = torch.stack([x0, y0], dim=0)  # shape: (2, H, W)
        coord = coord.unsqueeze(0).repeat(f_shape[0], 1, 1, 1)
        return coord

    def initialize_flow(self, fmap):
        coords0 = self.init_coord(fmap)
        coords1 = self.init_coord(fmap)

        return coords0, coords1

    def cvx_upsample(self, data, mask):
        N, C, H, W = data.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(data, [3, 3], padding=1)
        up_flow = up_flow.view(N, C, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, C, 8 * H, 8 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        if images.shape[1] == 2 and not self.has_shown_warning:
            print(
                "WARNING: splatflow is running in two-frame mode. Provide 3 input images to run in three-frame mode."
            )
            self.has_shown_warning = True

        flow_prs_01, mf_01, low_01 = self.forward_one_pair(images[:, 0], images[:, 1])
        if images.shape[1] > 2:
            mf_t = forward_warping(mf_01, low_01)
            flow_prs_12, mf_12, low_12 = self.forward_one_pair(
                images[:, 1], images[:, 2], mf_t=mf_t
            )
            out_flow = flow_prs_12[-1]
            flow_small = low_12
        else:
            out_flow = flow_prs_01[-1]
            flow_small = low_01

        out_flow = self.postprocess_predictions(out_flow, image_resizer, is_flow=True)
        outputs = {"flows": out_flow[:, None], "flow_small": flow_small}
        return outputs

    def forward_one_pair(self, image1, image2, mf_t=None):
        fmap1, fmap2 = self.fnet([image1, image2])

        corr_fn = CorrBlock(fmap1, fmap2, radius=4)

        coords0, coords1 = self.initialize_flow(fmap1)

        cnet = self.cnet(image1)
        net, inp = torch.split(cnet, [self.hdim, self.cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)
        atte_s = self.att(inp)

        flow_predictions = []

        for itr in range(self.iters):
            coords1 = coords1.detach()

            corr = corr_fn(coords1)

            flow = coords1 - coords0
            net, up_mask, delta_flow, mf = self.update(
                net, inp, corr, flow, atte_s, mf_t
            )
            coords1 = coords1 + delta_flow

            if (self.fast_inference and (itr == self.iters - 1)) or (
                not self.fast_inference
            ):
                flow_up = self.cvx_upsample(8 * (coords1 - coords0), up_mask)
                flow_predictions.append(flow_up)

            low = coords1 - coords0

        return flow_predictions, mf, low


@register_model
class splatflow(SplatFlow):
    pass



import torch
import torch.nn as nn
import torch.nn.functional as F
from .layer import ConvNextBlock


class FlowHead(nn.Module):
    def __init__(self, input_dim=128, hidden_dim=256, output_dim=4):
        super(FlowHead, self).__init__()
        self.conv1 = nn.Conv2d(input_dim, hidden_dim, 3, padding=1)
        self.conv2 = nn.Conv2d(hidden_dim, output_dim, 3, padding=1)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.conv2(self.relu(self.conv1(x)))


class BasicMotionEncoder(nn.Module):
    def __init__(self, corr_channel, dim=128):
        super(BasicMotionEncoder, self).__init__()
        cor_planes = corr_channel
        self.convc1 = nn.Conv2d(cor_planes, dim * 2, 1, padding=0)
        self.convc2 = nn.Conv2d(dim * 2, dim + dim // 2, 3, padding=1)
        self.convf1 = nn.Conv2d(2, dim, 7, padding=3)
        self.convf2 = nn.Conv2d(dim, dim // 2, 3, padding=1)
        self.conv = nn.Conv2d(dim * 2, dim - 2, 3, padding=1)

    def forward(self, flow, corr):
        cor = F.relu(self.convc1(corr))
        cor = F.relu(self.convc2(cor))
        flo = F.relu(self.convf1(flow))
        flo = F.relu(self.convf2(flo))

        cor_flo = torch.cat([cor, flo], dim=1)
        out = F.relu(self.conv(cor_flo))
        return torch.cat([out, flow], dim=1)


class BasicUpdateBlock(nn.Module):
    def __init__(self, corr_channel, num_blocks, hdim=128, cdim=128):
        # net: hdim, inp: cdim
        super(BasicUpdateBlock, self).__init__()
        self.encoder = BasicMotionEncoder(corr_channel=corr_channel, dim=cdim)
        self.refine = []
        for i in range(num_blocks):
            self.refine.append(ConvNextBlock(2 * cdim + hdim, hdim))
        self.refine = nn.ModuleList(self.refine)

    def forward(self, net, inp, corr, flow):
        motion_features = self.encoder(flow, corr)
        inp = torch.cat([inp, motion_features], dim=1)
        for blk in self.refine:
            net = blk(torch.cat([net, inp], dim=1))
        return net



from setuptools import setup
from torch.utils.cpp_extension import CppExtension, BuildExtension, CUDAExtension


setup(
    name="GANet",
    ext_modules=[
        CUDAExtension(
            "GANet",
            [
                "src/GANet_cuda.cpp",
                "src/GANet_kernel_share.cu",
                "src/NLF_kernel.cu",
            ],
        )
    ],
    cmdclass={"build_ext": BuildExtension},
)



import random
from typing import Optional, Sequence

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )
import torch
import torch.nn as nn

from ptlflow.utils.registry import register_model, trainable
from .pwc_modules import conv, upsample2d_as, rescale_flow, initialize_msra
from .pwc_modules import (
    WarpingLayer,
    FeatureExtractor,
    ContextNetwork,
    FlowEstimatorDense,
    OccContextNetwork,
    OccEstimatorDense,
)
from .irr_modules import OccUpsampleNetwork, RefineFlow, RefineOcc
from ..base_model.base_model import BaseModel
from .losses import MultiScaleEPE_PWC_Bi_Occ_upsample


class ScopeFlow(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scopeflow-chairs-ebfaa62d.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scopeflow-things-70e22d63.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scopeflow-kitti-a20c434d.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/scopeflow-sintel-21a91683.ckpt",
    }

    def __init__(
        self,
        div_flow: float = 0.05,
        search_range: int = 4,
        output_level: int = 4,
        num_levels: int = 7,
        num_chs: Sequence[int] = (3, 16, 32, 64, 96, 128, 196),
        dropout_p: float = 0.0,
        bilateral_mask: bool = False,
        random_freeze: bool = False,
        freeze_list: Optional[str] = None,
        seploss: bool = False,
        loss_perc: bool = False,
        train_batch_size: Optional[int] = None,
        **kwargs,
    ):
        super(ScopeFlow, self).__init__(
            output_stride=64,
            loss_fn=MultiScaleEPE_PWC_Bi_Occ_upsample(
                train_batch_size=train_batch_size, div_flow=div_flow
            ),
            **kwargs,
        )

        self.div_flow = div_flow
        self.search_range = search_range
        self.output_level = output_level
        self.num_levels = num_levels
        self.num_chs = num_chs
        self.dropout_p = dropout_p
        self.bilateral_mask = bilateral_mask
        self.random_freeze = random_freeze
        self.freeze_list = freeze_list
        self.seploss = seploss
        self.loss_perc = loss_perc
        self.train_batch_size = train_batch_size

        self.pwc_groups = [
            "extractor",
            "1x1",
            "flow_estimator",
            "context",
            "refine_flow",
            "attention",
        ] + ["occ"] * 3

        # Init sizes
        self.search_range = 4
        self.num_chs = [3, 16, 32, 64, 96, 128, 196]
        self.output_level = 4
        self.num_levels = 7
        self.ch_proj_size = 32
        self.min_layer = len(self.num_chs) - self.output_level - 1

        # Init generic layers
        self.leakyRELU = nn.LeakyReLU(0.1, inplace=True)
        self.feature_pyramid_extractor = FeatureExtractor(self.num_chs)
        self.warping_layer = WarpingLayer()
        self.correlation = SpatialCorrelationSampler(
            kernel_size=1, patch_size=2 * self.search_range + 1, padding=0
        )

        # Calc dimensions
        self.dim_corr = (self.search_range * 2 + 1) ** 2
        self.num_ch_in_flo = self.dim_corr + self.ch_proj_size + 2
        self.num_ch_in_occ = self.dim_corr + self.ch_proj_size + 1

        # Init PWC modules
        self.flow_estimators = FlowEstimatorDense(self.num_ch_in_flo)
        self.context_networks = ContextNetwork(self.num_ch_in_flo + 448 + 2)
        self.occ_estimators = OccEstimatorDense(self.num_ch_in_occ)
        self.occ_context_networks = OccContextNetwork(self.num_ch_in_occ + 448 + 1)
        self.occ_shuffle_upsample = OccUpsampleNetwork(11, 1)
        if dropout_p > 0.0:
            print("Starting dropout!")
            self.dropout = nn.Dropout2d(p=dropout_p, inplace=False)

        self.conv_1x1 = nn.ModuleList(
            [
                conv(
                    self.num_chs[ch],
                    self.ch_proj_size,
                    kernel_size=1,
                    stride=1,
                    dilation=1,
                )
                for ch in range(len(self.num_chs) - 1, self.min_layer, -1)
            ]
        )

        self.conv_1x1_1 = conv(
            self.num_chs[self.min_layer - 1], 3, kernel_size=1, stride=1, dilation=1
        )

        self.refine_flow = RefineFlow(2 + 1 + self.ch_proj_size, bilateral_mask)
        self.refine_occ = RefineOcc(1 + self.ch_proj_size + self.ch_proj_size)

        # Init weights
        initialize_msra(self.modules())
        self.param_groups = self._get_param_groups(self.pwc_groups)

        if freeze_list is not None:
            self.freezed_params = freeze_list.split(",")
            self._freeze()
            self.random_freeze = False
        self.mean_per_module = None

    def forward(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=0.0,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=False,
        )
        # Extract input
        x1_raw = images[:, 0]
        x2_raw = images[:, 1]
        batch_size, _, height_im, width_im = x1_raw.size()

        # Get pyramid, on the bottom level are original images
        cur_x1 = self.feature_pyramid_extractor(x1_raw)
        cur_x2 = self.feature_pyramid_extractor(x2_raw)
        if self.dropout_p > 0.0:
            # print("Dropout in ctx!")
            for xl in range(len(cur_x1)):
                # cur_x1[xl] = self.dropout(cur_x1[xl])
                cur_x2[xl] = self.dropout(cur_x2[xl])
        x1_pyramid = cur_x1 + [x1_raw]
        x2_pyramid = cur_x2 + [x2_raw]

        # Set output data structures
        output_dict = {}
        output_dict_eval = {}
        flows = []
        occs = []

        # Pre allocate output tensors
        flow_f, flow_b, occ_f, occ_b = self._allocate_out_tensors(
            x1_pyramid, batch_size
        )

        for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):
            if l <= self.output_level:
                # Warp
                x1_warp, x2_warp, flow_f, flow_b, occ_f, occ_b = self._warp(
                    l, x1, x2, flow_f, flow_b, occ_f, occ_b, height_im, width_im
                )

                # Correlate
                out_corr_relu_f = self._correlate(x1, x2_warp)
                out_corr_relu_b = self._correlate(x2, x1_warp)

                # Squash to channels projection size
                x1_1by1, x2_1by1 = self._squash(l, x1, x2)

                # Rescale Flow
                flow_f, flow_b = self._rescale_flow(flow_f, flow_b, height_im, width_im)

                # Estimate flow
                flow_cont_f, flow_cont_b = self._estimate_flow(
                    out_corr_relu_f, out_corr_relu_b, x1_1by1, x2_1by1, flow_f, flow_b
                )

                # Estimate occlusions
                occ_cont_f, occ_cont_b = self._estimate_occ(
                    out_corr_relu_f, out_corr_relu_b, x1_1by1, x2_1by1, occ_f, occ_b
                )

                # Prepare refinement inputs
                (
                    img1_resize,
                    img2_resize,
                    img1_warp,
                    img2_warp,
                    flow_cont_f,
                    flow_cont_b,
                ) = self._resize_and_warp_inputs(
                    x1_raw,
                    x2_raw,
                    flow_f,
                    flow_b,
                    flow_cont_f,
                    flow_cont_b,
                    height_im,
                    width_im,
                )

                # Refine Flow
                flow_f, flow_cont_f, flow_b, flow_cont_b = self._refine_flow(
                    flow_cont_f,
                    flow_cont_b,
                    img1_resize,
                    img2_resize,
                    img1_warp,
                    img2_warp,
                    x1_1by1,
                    x2_1by1,
                    height_im,
                    width_im,
                )

                # Refine Occlusions
                occ_f, occ_cont_f, occ_b, occ_cont_f = self._refine_occ(
                    x1_1by1,
                    x2_1by1,
                    flow_f,
                    flow_b,
                    occ_cont_f,
                    occ_cont_b,
                    height_im,
                    width_im,
                )

                # Collect layer's outputs
                flows.append([flow_cont_f, flow_cont_b, flow_f, flow_b])
                occs.append([occ_cont_f, occ_cont_b, occ_f, occ_b])

            else:
                # Final flow upsampling
                flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                flow_b = upsample2d_as(flow_b, x2, mode="bilinear")

                # Final occ upsampling
                occ_f, occ_b = self.occ_upsampling(
                    l, x1, x2, flow_f, flow_b, occ_f, occ_b, height_im, width_im
                )

                # Aggregate outputs
                flows.append([flow_f, flow_b])
                occs.append([occ_f, occ_b])

        flow_f_up = upsample2d_as(flow_f, x1_raw, mode="bilinear") * (
            1.0 / self.div_flow
        )
        flow_f_up = self.postprocess_predictions(flow_f_up, image_resizer, is_flow=True)
        flow_b_up = upsample2d_as(flow_b, x1_raw, mode="bilinear") * (
            1.0 / self.div_flow
        )
        flow_b_up = self.postprocess_predictions(flow_b_up, image_resizer, is_flow=True)
        occ_f_up = upsample2d_as(torch.sigmoid(occ_f), x1_raw, mode="bilinear")
        occ_f_up = self.postprocess_predictions(occ_f_up, image_resizer, is_flow=False)
        occ_b_up = upsample2d_as(torch.sigmoid(occ_b), x1_raw, mode="bilinear")
        occ_b_up = self.postprocess_predictions(occ_b_up, image_resizer, is_flow=False)

        outputs = {}
        if self.training:
            outputs["flow_preds"] = flows
            outputs["occ_preds"] = occs
            outputs["flows"] = flow_f_up[:, None]
            outputs["occs"] = occ_f_up[:, None]
            outputs["flows_b"] = flow_b_up[:, None]
            outputs["occs_b"] = occ_b_up[:, None]
        else:
            outputs["flows"] = flow_f_up[:, None]
            outputs["occs"] = occ_f_up[:, None]
            outputs["flows_b"] = flow_b_up[:, None]
            outputs["occs_b"] = occ_b_up[:, None]
        return outputs

    def submodules_summary(self):
        print(
            "Trainable submodules: {}".format(
                {p[0] for p in self.named_parameters() if p[1].requires_grad}
            )
        )
        if not self.mean_per_module:
            print("Initializing mean per module")
            self.mean_per_module = {
                p[0]: p[1].data.mean() for p in self.named_parameters()
            }
        else:
            changes = {
                p[0]: p[1].data.mean()
                for p in self.named_parameters()
                if p[1].data.mean() != self.mean_per_module[p[0]]
            }
            self.mean_per_module = {
                p[0]: p[1].data.mean() for p in self.named_parameters()
            }
            # print("Mean per submodule: {}".format(self.mean_per_module))
            print(f"Changes: {changes}")

    def freeze_random_weights(self):
        self.freezed_params = random.sample(
            self.pwc_groups, random.randrange(1, len(self.pwc_groups) / 2)
        )
        self._freeze(verify=True)

    def _freeze(self, verify=False):
        print(f"Freezing groups {self.freezed_params}")

        if verify:
            for param in self.named_parameters():
                if not param[1].requires_grad:
                    param[1].requires_grad = True

        fkeys = set()
        for freezing_group in self.freezed_params:
            for key, param in self.param_groups[freezing_group].items():
                param.requires_grad = False
                if param.grad is not None:
                    param.grad.data.zero_()
                    # print("Freezing {}, grad sum is {}".format(key, param.grad.sum()))
                fkeys.add(key)

        assert len(
            {p[0] for p in self.named_parameters() if not p[1].requires_grad}
        ) == len(fkeys)
        print(f"Freezed keys {fkeys}")
        print(
            "Params with grad {}".format(
                {p[0] for p in self.named_parameters() if p[1].requires_grad}
            )
        )

    def _get_param_groups(self, keys):
        pgroups = {}
        for substr in keys:
            pgroups[substr] = {k: t for k, t in self.named_parameters() if substr in k}

        # Verify each param has at least one group
        unique_group_params = set(
            [s for p in [list(pgroups[k].keys()) for k in keys] for s in p]
        )
        assert len(unique_group_params) == len(list(self.named_parameters()))

        return pgroups

    @staticmethod
    def _allocate_out_tensors(x1_pyramid, batch_size):
        (
            _,
            _,
            h_x1,
            w_x1,
        ) = x1_pyramid[0].size()
        flow_f = torch.zeros(batch_size, 2, h_x1, w_x1).to(
            dtype=x1_pyramid[0].dtype, device=x1_pyramid[0].device
        )
        flow_b = torch.zeros(batch_size, 2, h_x1, w_x1).to(
            dtype=x1_pyramid[0].dtype, device=x1_pyramid[0].device
        )
        occ_f = torch.zeros(batch_size, 1, h_x1, w_x1).to(
            dtype=x1_pyramid[0].dtype, device=x1_pyramid[0].device
        )
        occ_b = torch.zeros(batch_size, 1, h_x1, w_x1).to(
            dtype=x1_pyramid[0].dtype, device=x1_pyramid[0].device
        )
        return flow_f, flow_b, occ_f, occ_b

    def _warp(self, l, x1, x2, flow_f, flow_b, occ_f, occ_b, height_im, width_im):
        if l == 0:
            x2_warp = x2
            x1_warp = x1
        else:
            flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
            flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
            occ_f = upsample2d_as(occ_f, x1, mode="bilinear")
            occ_b = upsample2d_as(occ_b, x2, mode="bilinear")
            x2_warp = self.warping_layer(x2, flow_f, height_im, width_im, self.div_flow)
            x1_warp = self.warping_layer(x1, flow_b, height_im, width_im, self.div_flow)
        return x1_warp, x2_warp, flow_f, flow_b, occ_f, occ_b

    @staticmethod
    def collate_corr(corr, first_inp):
        b, ph, pw, h, w = corr.size()
        output_collated = corr.view(b, ph * pw, h, w)
        return output_collated / float(first_inp.size(1))

    def _correlate(self, x1, x2_warp):
        out_corr = self.correlation(x1, x2_warp)
        out_corr = out_corr.view(
            out_corr.shape[0], -1, out_corr.shape[3], out_corr.shape[4]
        )
        out_corr = out_corr / x1.shape[1]
        # out_corr = self.collate_corr(out_corr, x1)
        out_corr_relu = self.leakyRELU(out_corr)
        return out_corr_relu

    def _squash(self, l, x1, x2):
        if l != self.output_level:
            x1_1by1 = self.conv_1x1[l](x1)
            x2_1by1 = self.conv_1x1[l](x2)
        else:
            x1_1by1 = x1
            x2_1by1 = x2
        return x1_1by1, x2_1by1

    def _rescale_flow(self, flow_f, flow_b, height_im, width_im):
        flow_f = rescale_flow(flow_f, self.div_flow, width_im, height_im, to_local=True)
        flow_b = rescale_flow(flow_b, self.div_flow, width_im, height_im, to_local=True)
        return flow_f, flow_b

    def _estimate_flow(
        self, out_corr_relu_f, out_corr_relu_b, x1_1by1, x2_1by1, flow_f, flow_b
    ):
        # Estimate forward and backward residual flows
        x_intm_f, flow_res_f = self.flow_estimators(
            torch.cat([out_corr_relu_f, x1_1by1, flow_f], dim=1)
        )
        x_intm_b, flow_res_b = self.flow_estimators(
            torch.cat([out_corr_relu_b, x2_1by1, flow_b], dim=1)
        )

        # Add residual flow to the aggregated flow
        flow_est_f = flow_f + flow_res_f
        flow_est_b = flow_b + flow_res_b

        # Estimate residual context based refinement
        ctxf = self.context_networks(torch.cat([x_intm_f, flow_est_f], dim=1))
        ctxb = self.context_networks(torch.cat([x_intm_b, flow_est_b], dim=1))
        flow_cont_f = flow_est_f + ctxf
        flow_cont_b = flow_est_b + ctxb

        return flow_cont_f, flow_cont_b

    def _estimate_occ(
        self, out_corr_relu_f, out_corr_relu_b, x1_1by1, x2_1by1, occ_f, occ_b
    ):
        # Estimate residual occlusions
        x_intm_occ_f, occ_res_f = self.occ_estimators(
            torch.cat([out_corr_relu_f, x1_1by1, occ_f], dim=1)
        )
        x_intm_occ_b, occ_res_b = self.occ_estimators(
            torch.cat([out_corr_relu_b, x2_1by1, occ_b], dim=1)
        )

        # Add residual occlusions to the aggregated occlusions
        occ_est_f = occ_f + occ_res_f
        occ_est_b = occ_b + occ_res_b

        # Estimate residual context based refinement
        occ_cont_f = occ_est_f + self.occ_context_networks(
            torch.cat([x_intm_occ_f, occ_est_f], dim=1)
        )
        occ_cont_b = occ_est_b + self.occ_context_networks(
            torch.cat([x_intm_occ_b, occ_est_b], dim=1)
        )
        return occ_cont_f, occ_cont_b

    def _resize_and_warp_inputs(
        self,
        x1_raw,
        x2_raw,
        flow_f,
        flow_b,
        flow_cont_f,
        flow_cont_b,
        height_im,
        width_im,
    ):
        # Resizing input images for warping
        img1_resize = upsample2d_as(x1_raw, flow_f, mode="bilinear")
        img2_resize = upsample2d_as(x2_raw, flow_b, mode="bilinear")

        # Rescaling flow to the layer's size
        rescaled_f = rescale_flow(
            flow_cont_f, self.div_flow, width_im, height_im, to_local=False
        )
        rescaled_b = rescale_flow(
            flow_cont_b, self.div_flow, width_im, height_im, to_local=False
        )

        # Warping resized images
        img2_warp = self.warping_layer(
            img2_resize, rescaled_f, height_im, width_im, self.div_flow
        )
        img1_warp = self.warping_layer(
            img1_resize, rescaled_b, height_im, width_im, self.div_flow
        )

        return img1_resize, img2_resize, img1_warp, img2_warp, rescaled_f, rescaled_b

    def _refine_flow(
        self,
        flow_cont_f,
        flow_cont_b,
        img1_resize,
        img2_resize,
        img1_warp,
        img2_warp,
        x1_1by1,
        x2_1by1,
        height_im,
        width_im,
    ):
        flow_f = self.refine_flow(
            flow_cont_f.detach(), img1_resize - img2_warp, x1_1by1
        )
        flow_b = self.refine_flow(
            flow_cont_b.detach(), img2_resize - img1_warp, x2_1by1
        )

        flow_cont_f = rescale_flow(
            flow_cont_f, self.div_flow, width_im, height_im, to_local=False
        )
        flow_cont_b = rescale_flow(
            flow_cont_b, self.div_flow, width_im, height_im, to_local=False
        )
        flow_f = rescale_flow(
            flow_f, self.div_flow, width_im, height_im, to_local=False
        )
        flow_b = rescale_flow(
            flow_b, self.div_flow, width_im, height_im, to_local=False
        )
        return flow_f, flow_cont_f, flow_b, flow_cont_b

    def _refine_occ(
        self,
        x1_1by1,
        x2_1by1,
        flow_f,
        flow_b,
        occ_cont_f,
        occ_cont_b,
        height_im,
        width_im,
    ):
        x2_1by1_warp = self.warping_layer(
            x2_1by1, flow_f, height_im, width_im, self.div_flow
        )
        x1_1by1_warp = self.warping_layer(
            x1_1by1, flow_b, height_im, width_im, self.div_flow
        )

        occ_f = self.refine_occ(occ_cont_f.detach(), x1_1by1, x1_1by1 - x2_1by1_warp)
        occ_b = self.refine_occ(occ_cont_b.detach(), x2_1by1, x2_1by1 - x1_1by1_warp)
        return occ_f, occ_cont_f, occ_b, occ_cont_f

    def occ_upsampling(
        self, l, x1, x2, flow_f, flow_b, occ_f, occ_b, height_im, width_im
    ):
        x2_warp = self.warping_layer(x2, flow_f, height_im, width_im, self.div_flow)
        x1_warp = self.warping_layer(x1, flow_b, height_im, width_im, self.div_flow)
        flow_b_warp = self.warping_layer(
            flow_b, flow_f, height_im, width_im, self.div_flow
        )
        flow_f_warp = self.warping_layer(
            flow_f, flow_b, height_im, width_im, self.div_flow
        )

        if l != self.num_levels - 1:
            x1_in = self.conv_1x1_1(x1)
            x2_in = self.conv_1x1_1(x2)
            x1_w_in = self.conv_1x1_1(x1_warp)
            x2_w_in = self.conv_1x1_1(x2_warp)
        else:
            x1_in = x1
            x2_in = x2
            x1_w_in = x1_warp
            x2_w_in = x2_warp

        occ_f = self.occ_shuffle_upsample(
            occ_f, torch.cat([x1_in, x2_w_in, flow_f, flow_b_warp], dim=1)
        )
        occ_b = self.occ_shuffle_upsample(
            occ_b, torch.cat([x2_in, x1_w_in, flow_b, flow_f_warp], dim=1)
        )
        return occ_f, occ_b


@register_model
@trainable
class scopeflow(ScopeFlow):
    pass



from loguru import logger
import math
import torch
from torch import nn


class PositionEncodingSine(nn.Module):
    """
    This is a sinusoidal position encoding that generalized to 2-dimensional images
    """

    def __init__(self, d_model, max_shape=(256, 256)):
        """
        Args:
            max_shape (tuple): for 1/8 featmap, the max length of 256 corresponds to 2048 pixels
        """
        super().__init__()

        pe = torch.zeros((d_model, *max_shape))
        y_position = torch.ones(max_shape).cumsum(0).float().unsqueeze(0)
        x_position = torch.ones(max_shape).cumsum(1).float().unsqueeze(0)
        div_term = torch.exp(
            torch.arange(0, d_model // 2, 2).float()
            * (-math.log(10000.0) / d_model // 2)
        )
        div_term = div_term[:, None, None]  # [C//4, 1, 1]
        pe[0::4, :, :] = torch.sin(x_position * div_term)
        pe[1::4, :, :] = torch.cos(x_position * div_term)
        pe[2::4, :, :] = torch.sin(y_position * div_term)
        pe[3::4, :, :] = torch.cos(y_position * div_term)

        self.register_buffer("pe", pe.unsqueeze(0))  # [1, C, H, W]

    def forward(self, x):
        """
        Args:
            x: [N, C, H, W]
        """
        return x + self.pe[:, :, : x.size(2), : x.size(3)]


class LinearPositionEncoding(nn.Module):
    """
    This is a sinusoidal position encoding that generalized to 2-dimensional images
    """

    def __init__(self, d_model, max_shape=(256, 256)):
        """
        Args:
            max_shape (tuple): for 1/8 featmap, the max length of 256 corresponds to 2048 pixels
        """
        super().__init__()

        pe = torch.zeros((d_model, *max_shape))
        y_position = (
            torch.ones(max_shape).cumsum(0).float().unsqueeze(0) - 1
        ) / max_shape[0]
        x_position = (
            torch.ones(max_shape).cumsum(1).float().unsqueeze(0) - 1
        ) / max_shape[1]
        div_term = torch.arange(0, d_model // 2, 2).float()
        div_term = div_term[:, None, None]  # [C//4, 1, 1]
        pe[0::4, :, :] = torch.sin(x_position * div_term * math.pi)
        pe[1::4, :, :] = torch.cos(x_position * div_term * math.pi)
        pe[2::4, :, :] = torch.sin(y_position * div_term * math.pi)
        pe[3::4, :, :] = torch.cos(y_position * div_term * math.pi)

        self.register_buffer("pe", pe.unsqueeze(0), persistent=False)  # [1, C, H, W]

    def forward(self, x):
        """
        Args:
            x: [N, C, H, W]
        """
        # assert x.shape[2] == 80 and x.shape[3] == 80

        return x + self.pe[:, :, : x.size(2), : x.size(3)]


class LearnedPositionEncoding(nn.Module):
    """
    This is a sinusoidal position encoding that generalized to 2-dimensional images
    """

    def __init__(self, d_model, max_shape=(80, 80)):
        """
        Args:
            max_shape (tuple): for 1/8 featmap, the max length of 256 corresponds to 2048 pixels
        """
        super().__init__()

        self.pe = nn.Parameter(torch.randn(1, max_shape[0], max_shape[1], d_model))

    def forward(self, x):
        """
        Args:
            x: [N, C, H, W]
        """
        # assert x.shape[2] == 80 and x.shape[3] == 80

        return x + self.pe



from importlib.metadata import version

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable
from . import backbone
from . import transformer
from . import matching
from . import refine
from . import upsample
from . import utils
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exlude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        weights = [0.2, 1]
        for i in range(n_predictions):
            i_loss = (flow_preds[i] - flow_gt).abs()
            flow_loss += weights[i] * (valid * i_loss).mean()

        return flow_loss


class NeuFlow(BaseModel):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/neuflow-things-c402aa7a.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/neuflow-sintel-0d969ea2.ckpt",
    }

    def __init__(
        self,
        gamma: float = 0.8,
        max_flow: float = 400.0,
        feature_dim: int = 90,
        **kwargs,
    ) -> None:
        super().__init__(
            output_stride=16, loss_fn=SequenceLoss(gamma, max_flow), **kwargs
        )

        self.feature_dim = feature_dim

        if not version("torch").startswith("2"):
            raise ImportError(
                f"NeuFlow requires torch 2.X. Your current version is {version('torch')}"
            )

        self.backbone = backbone.CNNEncoder(feature_dim)
        self.cross_attn_s16 = transformer.FeatureAttention(
            feature_dim + 2,
            num_layers=2,
            bidir=True,
            ffn=True,
            ffn_dim_expansion=1,
            post_norm=True,
        )

        self.matching_s16 = matching.Matching()

        self.flow_attn_s16 = transformer.FlowAttention(feature_dim + 2)

        self.merge_s8 = torch.nn.Sequential(
            torch.nn.Conv2d(
                (feature_dim + 2) * 2,
                feature_dim * 2,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
            torch.nn.GELU(),
            torch.nn.Conv2d(
                feature_dim * 2,
                feature_dim,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=False,
            ),
        )

        self.refine_s8 = refine.Refine(feature_dim, patch_size=7, num_layers=6)

        self.conv_s8 = backbone.ConvBlock(
            3, feature_dim, kernel_size=8, stride=8, padding=0
        )

        self.upsample_s1 = upsample.UpSample(feature_dim, upsample_factor=8)

        for p in self.parameters():
            if p.dim() > 1:
                torch.nn.init.xavier_uniform_(p)

        self.curr_bhw = None

    def init_bhw(self, batch_size, height, width):
        self.backbone.init_pos_12(
            batch_size, height // 8, width // 8, dtype=self.dtype, device=self.device
        )
        self.matching_s16.init_grid(
            batch_size, height // 16, width // 16, dtype=self.dtype, device=self.device
        )

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""

        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=[-0.406, -0.456, -0.485],
            bgr_mult=[1 / 0.225, 1 / 0.224, 1 / 0.229],
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        img0 = images[:, 0]
        img1 = images[:, 1]

        bhw = (img0.shape[0], img0.shape[-2], img0.shape[-1])
        if bhw != self.curr_bhw:
            self.init_bhw(*bhw)
            self.curr_bhw = bhw

        feature0_s8, feature0_s16 = self.backbone(img0)
        feature1_s8, feature1_s16 = self.backbone(img1)

        feature0_s16, feature1_s16 = self.cross_attn_s16(feature0_s16, feature1_s16)
        flow0 = self.matching_s16.global_correlation_softmax(feature0_s16, feature1_s16)

        flow0 = self.flow_attn_s16(feature0_s16, flow0)

        feature0_s16 = F.interpolate(feature0_s16, scale_factor=2, mode="nearest")
        feature1_s16 = F.interpolate(feature1_s16, scale_factor=2, mode="nearest")

        feature0_s8 = self.merge_s8(torch.cat([feature0_s8, feature0_s16], dim=1))
        feature1_s8 = self.merge_s8(torch.cat([feature1_s8, feature1_s16], dim=1))

        flow0 = F.interpolate(flow0, scale_factor=2, mode="nearest") * 2

        delta_flow = self.refine_s8(
            feature0_s8, utils.flow_warp(feature1_s8, flow0), flow0
        )
        flow0 = flow0 + delta_flow

        flow_list = []
        if self.training:
            up_flow0 = (
                F.interpolate(
                    flow0, scale_factor=8, mode="bilinear", align_corners=True
                )
                * 8
            )
            up_flow0 = self.postprocess_predictions(
                up_flow0, image_resizer, is_flow=True
            )
            flow_list.append(up_flow0)

        feature0_s8 = self.conv_s8(img0)

        flow0 = self.upsample_s1(feature0_s8, flow0)
        flow0 = self.postprocess_predictions(flow0, image_resizer, is_flow=True)
        flow_list.append(flow0)

        if self.training:
            outputs = {"flows": flow0[:, None], "flow_preds": flow_list}
        else:
            outputs = {"flows": flow0[:, None]}

        return outputs


@register_model
@trainable
class neuflow(NeuFlow):
    pass



# =============================================================================
# Copyright 2023 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The following parts have been borrowed from other codes:
# ResidualBlock: https://github.com/princeton-vl/RAFT/blob/master/core/extractor.py
# forward_interpolate and bilinear_sampler: https://github.com/princeton-vl/RAFT/blob/master/core/utils/utils.py
# =============================================================================

from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F

from .pkconv import PKConv2d
from .local_timm.norm import GroupNorm, LayerNorm2d, BatchNorm2d


class SequentialPartial(nn.Module):
    def __init__(self, *ops) -> None:
        super().__init__()
        self.ops = nn.ModuleList(ops)

    def forward(self, x, out_ch=None):
        for op in self.ops:
            x = op(x, out_ch=out_ch)
        return x


class InterpolationTransition(nn.Module):
    """Change the input tensor shape using interpolation"""

    def __init__(
        self, use_zero_channel: bool = True, hw_scale_factor: float = 1.0
    ) -> None:
        super(InterpolationTransition, self).__init__()
        self.use_zero_channel = use_zero_channel
        self.hw_scale_factor = hw_scale_factor

    def forward(self, x, out_ch):
        b, c, h, w = x.shape
        if (self.hw_scale_factor < 0.999 or self.hw_scale_factor > 1.001) and (
            self.use_zero_channel or c == out_ch
        ):
            h, w = int(h * self.hw_scale_factor), int(w * self.hw_scale_factor)
            x = F.interpolate(x, (h, w), mode="bilinear", align_corners=True)

        if c != out_ch:
            if self.use_zero_channel:
                if c < out_ch:
                    x_zero = torch.zeros(
                        b, out_ch - c, h, w, dtype=x.dtype, device=x.device
                    )
                    x = torch.cat([x, x_zero], dim=1)
                elif c > out_ch:
                    x = x[:, :out_ch]
            else:
                h, w = int(h * self.hw_scale_factor), int(w * self.hw_scale_factor)
                x = F.interpolate(
                    x[None], (out_ch, h, w), mode="trilinear", align_corners=True
                )[0]
        return x


class ResidualPartialBlock(nn.Module):
    """ResNet block with PKConv layers"""

    def __init__(
        self,
        in_planes,
        planes,
        norm_layer=None,
        stride=1,
        use_out_activation=True,
        cache_pkconv_weights=False,
    ):
        super(ResidualPartialBlock, self).__init__()

        self.use_out_activation = use_out_activation
        self.conv1 = PKConv2d(
            in_planes,
            planes,
            kernel_size=3,
            padding=1,
            stride=stride,
            cache_weights=cache_pkconv_weights,
        )
        self.norm1 = (
            norm_layer(num_channels=planes)
            if norm_layer is not None
            else nn.Sequential()
        )
        self.conv2 = PKConv2d(
            planes,
            planes,
            kernel_size=3,
            padding=1,
            cache_weights=cache_pkconv_weights,
        )
        self.norm2 = (
            norm_layer(num_channels=planes)
            if norm_layer is not None
            else nn.Sequential()
        )
        self.act = nn.ReLU(inplace=True)

        if stride == 1:
            self.downsample = None
        else:
            self.downsample = PKConv2d(
                in_planes,
                planes,
                kernel_size=1,
                stride=stride,
                cache_weights=cache_pkconv_weights,
            )
            self.norm3 = (
                norm_layer(num_channels=planes)
                if norm_layer is not None
                else nn.Sequential()
            )

    def forward(self, x, out_ch):
        y = x
        y = self.act(self.norm1(self.conv1(y, out_ch)))
        y = self.norm2(self.conv2(y, out_ch))
        if self.use_out_activation:
            y = self.act(y)

        if self.downsample is not None:
            x = self.norm3(self.downsample(x, out_ch))

        out = x + y
        if self.use_out_activation:
            out = self.act(out)

        return out


class ResidualBlock(nn.Module):
    """ResNet block"""

    def __init__(
        self, in_planes, planes, norm_fn="group", stride=1, use_out_activation=True
    ):
        super(ResidualBlock, self).__init__()

        self.use_out_activation = use_out_activation

        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, padding=1, stride=stride
        )
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1)
        self.act = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes)
            self.norm2 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes)
            self.norm2 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            if not stride == 1:
                self.norm3 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm3
            )

    def forward(self, x):
        y = x
        y = self.act(self.norm1(self.conv1(y)))
        y = self.norm2(self.conv2(y))
        if self.use_out_activation:
            y = self.act(y)

        if self.downsample is not None:
            x = self.downsample(x)

        out = x + y
        if self.use_out_activation:
            out = self.act(out)

        return out


def bilinear_sampler(img, coords, mask=False):
    """Wrapper for grid_sample, uses pixel coordinates"""
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, align_corners=True)

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask

    return img


def get_norm_layer(layer_name, affine=False, num_groups=8):
    if layer_name == "batch":
        norm_layer = BatchNorm2d
    elif layer_name == "group":
        norm_layer = partial(
            GroupNorm,
            affine=affine,
            num_groups=num_groups,
        )
    elif layer_name == "layer":
        norm_layer = partial(LayerNorm2d, affine=affine)
    elif layer_name == "none":
        norm_layer = None
    else:
        raise ValueError(f"Unknown norm layer {layer_name}")
    return norm_layer



# =============================================================================
# Copyright 2023 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math
from typing import Optional, Sequence

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model, trainable, ptlflow_trained
from ptlflow.utils.utils import forward_interpolate_batch
from .pwc_modules import rescale_flow, upsample2d_as
from .update_partial import UpdatePartialBlock
from .corr import get_corr_block
from .pkconv_slk_encoder import PKConvSLKEncoder
from .utils import ResidualPartialBlock, InterpolationTransition, get_norm_layer
from .pkconv import PKConv2d
from ..base_model.base_model import BaseModel


class SequenceLoss(nn.Module):
    def __init__(self, gamma: float, max_flow: float):
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exclude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            pred = flow_preds[i]
            if (
                pred.shape[-2] != flow_gt.shape[-2]
                or pred.shape[-1] != flow_gt.shape[-1]
            ):
                pred = F.interpolate(
                    pred, size=flow_gt.shape[-2:], mode="bilinear", align_corners=True
                )
            i_weight = self.gamma ** (n_predictions - i - 1)
            diff = pred - flow_gt
            i_loss = (diff).abs()
            valid_loss = valid * i_loss
            flow_loss += i_weight * valid_loss.mean()

        return flow_loss


class UpNetPartial(nn.Module):
    def __init__(
        self,
        net_chs_fixed: int,
        enc_norm_type: str,
        use_norm_affine: bool,
        group_norm_num_groups: int,
        cache_pkconv_weights: bool,
    ) -> None:
        super().__init__()
        self.conv = PKConv2d(
            2 * net_chs_fixed,
            net_chs_fixed,
            1,
            cache_weights=cache_pkconv_weights,
        )
        self.act = nn.ReLU(inplace=True)
        self.res = ResidualPartialBlock(
            net_chs_fixed,
            net_chs_fixed,
            norm_layer=get_norm_layer(
                enc_norm_type,
                affine=use_norm_affine,
                num_groups=group_norm_num_groups,
            ),
            use_out_activation=False,
            cache_pkconv_weights=cache_pkconv_weights,
        )

    def forward(self, x):
        x = self.conv(x, x.shape[1] // 2)
        x = self.act(x)
        x = self.res(x, x.shape[1])
        return x


class RPKNet(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/rpknet-chairs-a705b345.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/rpknet-kitti-39504eb4.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/rpknet-sintel-e7cc969e.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/rpknet-things-f79b0d81.ckpt",
    }

    def __init__(
        self,
        pyramid_ranges: tuple[int, int] = (32, 8),
        iters: int = 12,
        input_pad_one_side: bool = False,
        input_bgr_to_rgb: bool = False,
        detach_flow: bool = True,
        corr_mode: str = "allpairs",
        upgate_norm_type: str = "group",
        use_norm_affine: bool = False,
        group_norm_num_groups: int = 8,
        corr_levels: int = 1,
        corr_range: int = 4,
        enc_norm_type: str = "group",
        enc_stem_stride: int = 2,
        enc_depth: int = 2,
        enc_mlp_ratio: float = 4.0,
        enc_hidden_chs: Sequence[int] = (32, 64, 96),
        enc_out_1x1_chs: str = "2.0",
        dec_gru_iters: int = 2,
        dec_gru_depth: int = 2,
        dec_gru_mlp_ratio: float = 4.0,
        dec_net_chs: Optional[int] = None,
        dec_inp_chs: Optional[int] = None,
        dec_motion_chs: int = 128,
        use_upsample_mask: bool = True,
        upmask_gradient_scale: float = 1.0,
        cache_pkconv_weights: bool = True,
        gamma: float = 0.8,
        max_flow: float = 400,
        **kwargs,
    ) -> None:
        for v in pyramid_ranges:
            assert (
                v > 0
            ), f"--pyramid_ranges values must be larger than 0, but found {v}"
            log_res = math.log2(v)
            assert (log_res) - int(
                log_res
            ) < 1e-3, f"--pyramid_ranges values must be powers of 2, but found {v}"
        num_recurrent_layers = int(math.log2(max(pyramid_ranges))) - 1
        output_stride = int(2 ** (num_recurrent_layers + 1))
        super().__init__(
            output_stride=output_stride,
            loss_fn=SequenceLoss(gamma, max_flow),
            **kwargs,
        )

        self.pyramid_ranges = pyramid_ranges
        self.iters = iters
        self.input_pad_one_side = input_pad_one_side
        self.input_bgr_to_rgb = input_bgr_to_rgb
        self.detach_flow = detach_flow
        self.corr_mode = corr_mode
        self.upgate_norm_type = upgate_norm_type
        self.use_norm_affine = use_norm_affine
        self.group_norm_num_groups = group_norm_num_groups
        self.corr_levels = corr_levels
        self.corr_range = corr_range
        self.enc_norm_type = enc_norm_type
        self.enc_stem_stride = enc_stem_stride
        self.enc_depth = enc_depth
        self.enc_mlp_ratio = enc_mlp_ratio
        self.enc_hidden_chs = enc_hidden_chs
        self.enc_out_1x1_chs = enc_out_1x1_chs
        self.dec_gru_iters = dec_gru_iters
        self.dec_gru_depth = dec_gru_depth
        self.dec_gru_mlp_ratio = dec_gru_mlp_ratio
        self.dec_net_chs = dec_net_chs
        self.dec_inp_chs = dec_inp_chs
        self.dec_motion_chs = dec_motion_chs
        self.use_upsample_mask = use_upsample_mask
        self.upmask_gradient_scale = upmask_gradient_scale
        self.cache_pkconv_weights = cache_pkconv_weights

        if isinstance(self.enc_out_1x1_chs, str):
            self.enc_out_1x1_chs = (
                float(self.enc_out_1x1_chs)
                if "." in self.enc_out_1x1_chs
                else int(self.enc_out_1x1_chs)
            )

        if isinstance(self.enc_out_1x1_chs, float):
            self.out_1x1_factor = self.enc_out_1x1_chs
            self.out_1x1_abs_chs = int(self.enc_out_1x1_chs * self.enc_hidden_chs[-1])
        else:
            self.out_1x1_factor = None
            self.out_1x1_abs_chs = self.enc_out_1x1_chs

        self.max_feat_chs = max(
            self.enc_hidden_chs[-1],
            self.out_1x1_abs_chs,
        )

        net_chs = self.dec_net_chs
        inp_chs = self.dec_inp_chs
        if net_chs is None or inp_chs is None:
            base_chs = self.out_1x1_abs_chs
            if base_chs < 1:
                base_chs = enc_hidden_chs[-1]

            base_chs = base_chs // 3 * 2

            if net_chs is None and inp_chs is None:
                net_chs = inp_chs = base_chs // 2
            elif net_chs is None and inp_chs is not None:
                net_chs = base_chs - inp_chs
            elif net_chs is not None and inp_chs is None:
                inp_chs = base_chs - net_chs
        self.net_chs_fixed = net_chs
        self.inp_chs_fixed = inp_chs

        self.pyramid_levels = [
            num_recurrent_layers + 1 - int(math.log2(v)) for v in self.pyramid_ranges
        ]
        self.min_pyr_level = min(self.pyramid_levels)
        self.max_pyr_level = max(self.pyramid_levels)

        pyr_range = [min(self.pyramid_ranges), max(self.pyramid_ranges)]

        self.fnet = PKConvSLKEncoder(
            pyr_range=pyr_range,
            hidden_chs=self.enc_hidden_chs,
            out_1x1_abs_chs=self.out_1x1_abs_chs,
            out_1x1_factor=self.out_1x1_factor,
            mlp_ratio=self.enc_mlp_ratio,
            depth=self.enc_depth,
            norm_layer=get_norm_layer(
                self.enc_norm_type,
                affine=self.use_norm_affine,
                num_groups=self.group_norm_num_groups,
            ),
            stem_stride=self.enc_stem_stride,
            cache_pkconv_weights=self.cache_pkconv_weights,
        )

        self.dim_corr = (self.corr_range * 2 + 1) ** 2 * self.corr_levels

        self.update_block = UpdatePartialBlock(
            pyramid_ranges=self.pyramid_ranges,
            corr_levels=self.corr_levels,
            corr_range=self.corr_range,
            net_chs_fixed=self.net_chs_fixed,
            inp_chs_fixed=self.inp_chs_fixed,
            group_norm_num_groups=self.group_norm_num_groups,
            use_norm_affine=self.use_norm_affine,
            dec_motion_chs=self.dec_motion_chs,
            dec_gru_depth=self.dec_gru_depth,
            dec_gru_iters=self.dec_gru_iters,
            dec_gru_mlp_ratio=self.dec_gru_mlp_ratio,
            use_upsample_mask=self.use_upsample_mask,
            upmask_gradient_scale=self.upmask_gradient_scale,
            cache_pkconv_weights=self.cache_pkconv_weights,
        )

        self.upnet_layer = InterpolationTransition(False, 2)

        self.upnet_gate_layer = UpNetPartial(
            net_chs_fixed=self.net_chs_fixed,
            enc_norm_type=enc_norm_type,
            use_norm_affine=use_norm_affine,
            group_norm_num_groups=group_norm_num_groups,
            cache_pkconv_weights=cache_pkconv_weights,
        )

        self.input_act = nn.ReLU()

        self.has_trained_on_ptlflow = True

    def coords_grid(self, batch, ht, wd):
        coords = torch.meshgrid(
            torch.arange(ht, dtype=self.dtype, device=self.device),
            torch.arange(wd, dtype=self.dtype, device=self.device),
            indexing="ij",
        )
        coords = torch.stack(coords[::-1], dim=0)
        return coords[None].repeat(batch, 1, 1, 1)

    def upsample_flow(self, flow, mask, factor, ch=2):
        """Upsample flow field [H/f, W/f, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, factor, factor, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, ch, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, ch, factor * H, factor * W)

    def forward(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=self.input_bgr_to_rgb,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=(not self.input_pad_one_side),
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        flow_init = None
        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            flow_init = inputs["prev_preds"]["flow_small"]

        flow_predictions, flow_small, flow_up = self.predict(image1, image2, flow_init)
        flow_up = self.postprocess_predictions(flow_up, image_resizer, is_flow=True)
        for i, p in enumerate(flow_predictions):
            flow_predictions[i] = self.postprocess_predictions(
                p, image_resizer, is_flow=True
            )
        outputs = {"flows": flow_up[:, None], "flow_small": flow_small}

        if self.training:
            outputs["flow_preds"] = flow_predictions

        return outputs

    def predict(self, x1_raw, x2_raw, flow_init):
        b, _, height_im, width_im = x1_raw.size()

        x1_pyramid = self.fnet(x1_raw)
        x2_pyramid = self.fnet(x2_raw)

        # outputs
        flows = []

        pred_stride = min(self.pyramid_ranges)

        ipyr = 0
        if self.training:
            ipyr = self.global_step % (len(self.pyramid_ranges) // 2)
        pyr_levels = self.pyramid_levels[2 * ipyr : 2 * ipyr + 2]

        start_level, output_level = pyr_levels
        pass_pyramid1 = x1_pyramid[start_level : output_level + 1]
        pass_pyramid2 = x2_pyramid[start_level : output_level + 1]

        level_diff = output_level - start_level
        iters_per_level = [int(math.ceil(float(self.iters) / (level_diff + 1)))] * (
            level_diff + 1
        )

        # init
        (
            b_size,
            _,
            h_x1,
            w_x1,
        ) = pass_pyramid1[0].size()

        if flow_init is not None:
            flow = forward_interpolate_batch(flow_init)
        else:
            flow = torch.zeros(
                b_size,
                2,
                h_x1,
                w_x1,
                dtype=pass_pyramid1[0].dtype,
                device=pass_pyramid1[0].device,
            )

        net = None
        for l, (x1, x2) in enumerate(zip(pass_pyramid1, pass_pyramid2)):
            # Split feature channels into matching (x) and context (c)
            xh = x1.shape[1]
            ch = xh // 3
            x1, cn1 = torch.split(x1, [xh - ch, ch], dim=1)
            x2, cn2 = torch.split(x2, [xh - ch, ch], dim=1)
            halfch = ch // 2
            i1, n1 = torch.split(cn1, [ch - halfch, halfch], dim=1)
            i2, n2 = torch.split(cn2, [ch - halfch, halfch], dim=1)
            inp = torch.cat([i1, i2], 1)
            inp = self.input_act(inp)
            net_tmp = torch.cat([n1, n2], 1)

            coords0 = self.coords_grid(x1.shape[0], x1.shape[2], x1.shape[3])

            corr_fn = get_corr_block(
                x1,
                x2,
                self.corr_levels,
                self.corr_range,
                alternate_corr=self.corr_mode == "local",
            )

            if net is None:
                net = torch.tanh(net_tmp)
            else:
                net = self.upnet_layer(net, net_tmp.shape[1])
                net = torch.tanh(net)

                net_skip = torch.tanh(net_tmp)
                gate = torch.sigmoid(
                    self.upnet_gate_layer(torch.cat([net, net_skip], dim=1))
                )
                net = gate * net + (1.0 - gate) * net_skip

            if l > 0:
                flow = rescale_flow(flow, x1.shape[-1], x1.shape[-2], to_local=False)
                flow = upsample2d_as(flow, x1, mode="bilinear")

            for _ in range(iters_per_level[l]):
                if self.detach_flow:
                    flow = flow.detach()

                # correlation
                out_corr = corr_fn(coords0 + flow)

                flow_res, net, mask = self.update_block(net, inp, out_corr, flow)

                flow = flow + flow_res

                out_flow = flow
                small_flow = out_flow
                out_flow = rescale_flow(out_flow, width_im, height_im, to_local=False)

                if l < (output_level - start_level) or mask is None:
                    out_flow = upsample2d_as(out_flow, x1_raw, mode="bilinear")
                else:
                    out_flow = self.upsample_flow(out_flow, mask, pred_stride)

                flows.append(out_flow)

        small_flow = rescale_flow(
            small_flow,
            pass_pyramid1[0].shape[-1],
            pass_pyramid1[0].shape[-2],
            to_local=False,
        )
        small_flow = upsample2d_as(small_flow, pass_pyramid1[0], mode="bilinear")

        return flows, small_flow, out_flow


@register_model
@trainable
@ptlflow_trained
class rpknet(RPKNet):
    pass
